indexing
	description: "Integers of arbitrary size"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	GMP_INTEGER_REF

inherit
	COMPARABLE
		redefine
			copy, default_create, infix ">", is_equal, out
		end

	GMP_INTEGER_FUNCTIONS
		export
			{NONE} all
		redefine
			copy, default_create, is_equal, out
		end

	MEMORY
		export
			{NONE} all
		redefine
			copy, default_create, dispose, is_equal, out
		end

	MEMORY_STRUCTURE
		rename
			make as memory_structure_make
		export
			{NONE} make_by_pointer, memory_structure_make
		redefine
			copy, default_create, is_equal, out
		end

	NUMERIC
		rename
			infix "/" as infix "//"
		redefine
			copy, default_create, is_equal, out
		end

create
	default_create, make, make_by_pointer, make_from_gmp_integer,
	make_from_gmp_rational, make_from_gmp_real, make_from_string

feature {NONE} -- Initialization

	default_create is
			-- Initialize `Current'.
		do
			memory_structure_make
			mpz_init (item)
		end

	make (i: INTEGER) is
			-- Initialize `Current' from `i'.
		do
			memory_structure_make
			mpz_init_set_si (item, i)
		end

	make_from_gmp_integer (i: like Current) is
			-- Initialize `Current' from `i'.
		do
			memory_structure_make
			mpz_init_set (item, i.item)
		end

	make_from_gmp_rational (r: GMP_RATIONAL_REF) is
			-- Initialize `Current' from `r'.
		do
			memory_structure_make
			mpz_set_q (item, r.item)
		end

	make_from_gmp_real (r: GMP_REAL_REF) is
			-- Initialize `Current' from `r'.
		do
			memory_structure_make
			mpz_set_f (item, r.item)
		end

	make_from_string (s: STRING) is
			-- Initialize `Current' from `s'.
		require
			s_not_void: s /= Void
			s_is_integer: s.is_integer
		local
			a: ANY
		do
			memory_structure_make
			a := s.to_c
			mpz_init_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Decimal_base)
		end

feature -- Access

	one: like Current is
			-- Neutral element for "*" and "/"
		do
			create Result.make (1)
		end

	sign: INTEGER is
			-- Sign value (0, -1 or 1)
		do
			Result := mpz_sgn (item)
		ensure
			three_way: Result = three_way_comparison (zero)
		end

	zero: like Current is
			-- Neutral element for "+" and "-"
		do
			create Result
		end

feature -- Measurement

	precision: INTEGER is
			-- Number of bits used.
		do
			Result := mp_size (item).abs * (feature {GMP_GLOBAL_CONSTANTS}.Gmp_limb_bits)
		end

	structure_size: INTEGER is
			-- Size to allocate (in bytes)
		do
			Result := c_structure_size
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
		do
			Result := not other.is_equal (zero)
		end

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		local
			int_8_ref: INTEGER_8_REF
			int_16_ref: INTEGER_16_REF
			int_ref: INTEGER_REF
		do
			int_ref ?= other
			if int_ref /= Void then
				Result := int_ref.item >= 0
			else
				int_16_ref ?= other
				if int_16_ref /= Void then
					Result := int_16_ref.to_integer >= 0
				else
					int_8_ref ?= other
					if int_8_ref /= Void then
						Result := int_8_ref.to_integer >= 0
					end
				end
			end
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN is
			-- Is current object less than `other'?
		do
			Result := mpz_cmp (item, other.item) < 0
		end

	infix ">" (other: like Current): BOOLEAN is
			-- Is current object greater than `other'?
		do
			Result := mpz_cmp (item, other.item) > 0
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := mpz_cmp (item, other.item) = 0
		end

feature -- Element change

	set_from_binary_string (b: STRING) is
			-- Set value from `b'
			-- `b' is assumed to be a binary representation.
		require
			b_exists: b /= Void
			b_not_empty: not b.is_empty
		local
			a: ANY
		do
			a := b.to_c
			mpz_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Binary_base)
		end

	set_from_double (d: DOUBLE) is
			-- Set value from `d'
		do
			mpz_set_d (item, d)
		end

	set_from_gmp_rational (r: GMP_RATIONAL_REF) is
			-- Set value from `r'
		require
			r_exists: r /= Void
		do
			mpz_set_q (item, r.item)
		end

	set_from_gmp_real (r: GMP_REAL_REF) is
			-- Set value from `r'
		require
			r_exists: r /= Void
		do
			mpz_set_f (item, r.item)
		end

	set_from_hex_string (h: STRING) is
			-- Set value from `h'
			-- `h' is assumed to be a hexadecimal representation.
		require
			h_exists: h /= Void
			h_not_empty: not h.is_empty
		local
			a: ANY
		do
			a := h.to_c
			mpz_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Hexadecimal_base)
		end

	set_from_integer (i: INTEGER) IS
			-- Set value from `i'
		do
			mpz_set_si (item, i)
		end

	set_from_string (s: STRING) is
			-- Set value from `s'
			-- `s' is assumed to be a decimal representation.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
		do
			a := s.to_c
			mpz_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Decimal_base)
		end

feature {NONE} -- Removal

	dispose is
			-- Action to be executed just before garbage collection
			-- reclaims an object.
		do
			mpz_clear (item)
		end

feature -- Conversion

	to_binary_string: STRING is
			-- Convert to a hexadecimal string
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (item, feature {GMP_BASE_CONSTANTS}.Binary_base) + String_extra_bytes)
			mpz_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Binary_base, item)
			Result := s.string
		end

	to_double: DOUBLE is
			-- Convert to a DOUBLE
		do
			Result := mpz_get_d (item) 
		end

	to_gmp_rational: GMP_RATIONAL_REF is
			-- Convert to a GMP_RATIONAL_REF.
		do
			create Result.make_from_gmp_integer (Current)
		end

	to_gmp_real: GMP_REAL_REF is
			-- Convert to a GMP_REAL_REF.
		do
			create Result.make_from_gmp_integer (Current)
		end

	to_hex_string: STRING is
			-- Convert to a hexadecimal string
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (item, feature {GMP_BASE_CONSTANTS}.Hexadecimal_base) + String_extra_bytes)
			mpz_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Hexadecimal_base, item)
			Result := s.string
		end

	to_integer: INTEGER is
			-- Convert to an INTEGER
		do
			Result := mpz_get_si (item) 
		end

	to_mpfr: MPFR_REF is
			-- Convert to an MPFR_REF.
		do
			Result := to_gmp_real.to_mpfr
		end

	to_octal_string: STRING is
			-- Convert to a hexadecimal string
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (item, feature {GMP_BASE_CONSTANTS}.Octal_base) + String_extra_bytes)
			mpz_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Octal_base, item)
			Result := s.string
		end

	to_string_in_base (a_base: INTEGER): STRING is
			-- Convert to a string in base `a_base'
		require
			a_base_in_range: (2 <= a_base) and (a_base <= 36)
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (item, a_base) + String_extra_bytes)
			mpz_get_str (s.item, a_base, item)
			Result := s.string
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
		do
			mpz_set (item, other.item)
		end

feature -- Basic operations

	abs: like Current is
			-- Absolute value
		do
			create Result.make_from_gmp_integer (Current)
			mpz_abs (Result.item, Result.item)
		ensure
			non_negative: Result >= zero
			same_absolute_value: (Result.is_equal (Current)) or (Result.is_equal (-Current))
		end

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		do
			create Result
			mpz_add (Result.item, item, other.item)
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'
		do
			create Result
			mpz_sub (Result.item, item, other.item)
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'
		do
			create Result
			mpz_mul (Result.item, item, other.item)
		end

	infix "/" (other: like Current): GMP_REAL_REF is
			-- Division by `other'
		require
			other_exists: other /= Void
			good_divisor: divisible (other)
		do
			create Result.make_from_gmp_real (create {GMP_REAL_REF}.make_from_gmp_integer (Current) /
			                                  create {GMP_REAL_REF}.make_from_gmp_integer (other))
		end

	infix "//" (other: like Current): like Current is
			-- Division by `other'
		do
			create Result
			mpz_tdiv_q (Result.item, item, other.item)
		end

	infix "\\" (other: like Current): like Current is
			-- Remainder after division by `other'
		require
			other_exists: other /= Void
			good_divisor: divisible (other)
		do
			create Result
			mpz_tdiv_r (Result.item, item, other.item)
		end

	infix "^" (other: NUMERIC): like Current is
			-- Integer power of Current by `other'
		require
			other_exists: other /= Void
			good_exponent: exponentiable (other)
		local
			int_8_ref: INTEGER_8_REF
			int_16_ref: INTEGER_16_REF
			int_ref: INTEGER_REF
		do
			int_ref ?= other
			if int_ref /= Void then
				create Result
				mpz_pow_ui (Result.item, item, int_ref.item)
			else
				int_16_ref ?= other
				if int_16_ref /= Void then
					create Result
					mpz_pow_ui (Result.item, item, int_16_ref.to_integer)
				else
					int_8_ref ?= other
					if int_8_ref /= Void then
						create Result
						mpz_pow_ui (Result.item, item, int_8_ref.to_integer)
					end
				end
			end
		end

	prefix "+": like Current is
			-- Unary plus
		do
			create Result.make_from_gmp_integer (Current)
		end

	prefix "-": like Current is
			-- Unary minus
		do
			create Result
			mpz_neg (Result.item, item) 
		end

feature -- Bit operations

	infix "&", bit_and (i: like Current): like Current is
			-- Bitwise and between Current' and `i'.
		require
			i_not_void: i /= Void
		do
			create Result
			mpz_and (Result.item, item, i.item)
		ensure
			bitwise_and_not_void: Result /= Void
		end

	infix "|", bit_or (i: like Current): like Current is
			-- Bitwise or between Current' and `i'.
		require
			i_not_void: i /= Void
		do
			create Result
			mpz_ior (Result.item, item, i.item) 
		ensure
			bitwise_or_not_void: Result /= Void
		end

	bit_xor (i: like Current): like Current is
			-- Bitwise xor between Current' and `i'.
		require
			i_not_void: i /= Void
		do
			create Result
			mpz_xor (Result.item, item, i.item)
		ensure
			bitwise_xor_not_void: Result /= Void
		end

	bit_not: like Current is
			-- One's complement of Current.
		do
			create Result
			mpz_com (Result.item, item)
		ensure
			bit_not_not_void: Result /= Void
		end

	infix "|<<", bit_shift_left (n: INTEGER): like Current is
			-- Shift Current from `n' position to left.
		require
			n_nonnegative: n >= 0
		do
			create Result
			mpz_mul_2exp (Result.item, item, n) 
		ensure
			bit_shift_left_not_void: Result /= Void
		end

	bit_test (n: INTEGER): BOOLEAN is
			-- Test `n'-th position of Current.
		require
			n_nonnegative: n >= 0
		do
			Result := mpz_tstbit (item, n) = 1
		end

	set_bit (b: BOOLEAN; n: INTEGER): like Current is
			-- Copy of current with `n'-th position
			-- set to 1 if `b', 0 otherwise.
		require
			n_nonnegative: n >= 0
		do
			create Result.make_from_gmp_integer (Current)
			if b then
				mpz_setbit (Result.item, n) 
			else
				mpz_clrbit (Result.item, n)
			end
		end

feature -- Output

	out: STRING is
			-- Simple string representation.
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (item, feature {GMP_BASE_CONSTANTS}.Decimal_base) + String_extra_bytes)
			mpz_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Decimal_base, item)
			Result := s.string
		end

feature {NONE} -- Implementation

	String_extra_bytes: INTEGER is 2

feature {NONE} -- Externals

	c_structure_size: INTEGER is
			-- Size of  "C" struct mpz_t.
		external
			"C [macro %"gmp.h%"]"
		alias
			"sizeof (__mpz_struct)"
		end

	mp_size (p: POINTER): INTEGER is
		external
			"C [struct %"gmp.h%"] (__mpz_struct): EIF_INTEGER"
		alias
			"_mp_size"
		end

end -- class GMP_INTEGER_REF
