indexing
	description: "Rational numbers"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	GMP_RATIONAL_REF

inherit
	COMPARABLE
		redefine
			copy, default_create, infix ">", is_equal, out
		end

	GMP_INTEGER_FUNCTIONS
		export
			{NONE} all
		redefine
			copy, default_create, is_equal, out
		end

	GMP_RATIONAL_FUNCTIONS
		export
			{NONE} all
		redefine
			copy, default_create, is_equal, out
		end

	MEMORY
		export
			{NONE} all
		redefine
			copy, default_create, dispose, is_equal, out
		end

	MEMORY_STRUCTURE
		rename
			make as memory_structure_make
		export
			{NONE} make_by_pointer, memory_structure_make
		redefine
			copy, default_create, is_equal, out
		end

	NUMERIC
		redefine
			copy, default_create, is_equal, out
		end

create
	default_create, make, make_from_double, make_from_gmp_integer,
	make_from_gmp_rational, make_from_gmp_real, make_from_integer,
	make_from_string

feature {NONE} -- Initialization

	default_create is
			-- Initialize `Current'.
		do
			memory_structure_make
			mpq_init (item)
		end

	make (n, d: GMP_INTEGER_REF) is
			-- Initialize `numerator' from `n' and `denominator' from `d'.
		require
			n_exists: n /= Void
			d_exists: d /= Void
			d_not_zero: not d.is_equal (d.zero)
		do
			memory_structure_make
			mpq_init (item)
			mpq_set_num (item, n.item)
			mpq_set_den (item, d.item)
			mpq_canonicalize (item)
		end

	make_from_double (d: DOUBLE) is
			-- Initialize from `d'.
		do
			memory_structure_make
			mpq_init (item)
			mpq_set_d (item, d)
			mpq_canonicalize (item)
		end

	make_from_gmp_integer (i: GMP_INTEGER_REF) is
			-- Initialize from `i'.
		require
			i_exists: i /= Void
		do
			memory_structure_make
			mpq_init (item)
			mpq_set_z (item, i.item)
		end

	make_from_gmp_rational (r: GMP_RATIONAL_REF) is
			-- Initialize from `r'.
		require
			r_exists: r /= Void
		do
			memory_structure_make
			mpq_init (item)
			mpq_set (item, r.item)
		end

	make_from_gmp_real (r: GMP_REAL_REF) is
			-- Initialize from `r'.
		require
			r_exists: r /= Void
		do
			memory_structure_make
			mpq_init (item)
			mpq_set_f (item, r.item)
			mpq_canonicalize (item)
		end

	make_from_integer (n, d: INTEGER) is
			-- Initialize `numerator' from `n' and `denominator' from `d'.
		require
			d_positive: d > 0
		do
			memory_structure_make
			mpq_init (item)
			mpq_set_si (item, n, d)
			mpq_canonicalize (item)
		end

	make_from_string (s: STRING) is
			-- Initialize from `s'.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
		do
			memory_structure_make
			mpq_init (item)
			a := s.to_c
			mpq_set_str (item, $a, 10)
			mpq_canonicalize (item)
		end

feature -- Access

	denominator: GMP_INTEGER_REF is
			-- Denominator
		do
			create Result
			mpq_get_den (Result.item, item)
		end

	numerator: GMP_INTEGER_REF is
			-- Numerator
		do
			create Result
			mpq_get_num (Result.item, item)
		end

	one: like Current is
			-- Neutral element for "*" and "/"
		do
			create Result.make_from_integer (1, 1)
		end

	sign: INTEGER is
			-- Sign value (0, -1 or 1)
		do
			Result := mpq_sgn (item)
		ensure
			three_way: Result = three_way_comparison (zero)
		end

	zero: like Current is
			-- Neutral element for "+" and "-"
		do
			create Result.make_from_integer (0, 1)
		end

feature -- Measurement

	structure_size: INTEGER is
			-- Size to allocate (in bytes)
		do
			Result := c_structure_size
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
		do
			Result := not other.is_equal (zero)
		end

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		local
			int_8_ref: INTEGER_8_REF
			int_16_ref: INTEGER_16_REF
			int_ref: INTEGER_REF
		do
			int_ref ?= other
			if int_ref /= Void then
				Result := (int_ref.item >= 0) or else (not is_equal (zero))
			else
				int_16_ref ?= other
				if int_16_ref /= Void then
					Result := (int_16_ref.to_integer >= 0) or else (not is_equal (zero))
				else
					int_8_ref ?= other
					if int_8_ref /= Void then
						Result := (int_8_ref.to_integer >= 0) or else (not is_equal (zero))
					end
				end
			end
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN is
			-- Is current object less than `other'?
		do
			Result := mpq_cmp (item, other.item) < 0
		end

	infix ">" (other: like Current): BOOLEAN is
			-- Is current object greater than `other'?
		do
			Result := mpq_cmp (item, other.item) > 0
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := mpq_equal (item, other.item) /= 0
		end

feature -- Element change

	set (n, d: GMP_INTEGER_REF) is
			-- Set `numerator' from `n' and `denominator' from `d'.
		require
			n_exists: n /= Void
			d_exists: d /= Void
			d_not_zero: not d.is_equal (d.zero)
		do
			mpq_set_num (item, n.item)
			mpq_set_den (item, d.item)
			mpq_canonicalize (item)
		end

	set_from_double (d: DOUBLE) is
			-- Set from `d'.
		do
			mpq_set_d (item, d)
			mpq_canonicalize (item)
		end

	set_from_gmp_integer (i: GMP_INTEGER_REF) is
			-- Set from `i'.
		require
			i_exists: i /= Void
		do
			mpq_set_z (item, i.item)
		end

	set_from_gmp_rational (r: GMP_RATIONAL_REF) is
			-- Set from `r'.
		require
			r_exists: r /= Void
		do
			mpq_set (item, r.item)
		end

	set_from_gmp_real (r: GMP_REAL_REF) is
			-- Set from `r'.
		require
			r_exists: r /= Void
		do
			mpq_set_f (item, r.item)
			mpq_canonicalize (item)
		end

	set_from_integer (n, d: INTEGER) is
			-- Set `numerator' from `n' and `denominator' from `d'.
		require
			d_positive: d > 0
		do
			mpq_set_si (item, n, d)
			mpq_canonicalize (item)
		end

	set_from_string (s: STRING) is
			-- Set from `s'.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
		do
			a := s.to_c
			mpq_set_str (item, $a, 10)
			mpq_canonicalize (item)
		end

feature -- Removal

	dispose is
			-- Action to be executed just before garbage collection
			-- reclaims an object.
		do
			mpq_clear (item)
		end

feature -- Conversion

	to_binary_string: STRING is
			-- Convert to a hexadecimal string
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (numerator.item, feature {GMP_BASE_CONSTANTS}.Binary_base) +
			                     mpz_sizeinbase (denominator.item, feature {GMP_BASE_CONSTANTS}.Binary_base) +
			                     String_extra_bytes)
			mpq_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Binary_base, item)
			Result := s.string
		end

	to_double: DOUBLE is
			-- Convert to a DOUBLE
		do
			Result := mpq_get_d (item) 
		end

	to_gmp_integer: GMP_INTEGER_REF is
			-- Comvert to a GMP_INTEGER_REF.
		do
			Result := numerator // denominator
		end

	to_gmp_real: GMP_REAL_REF is
			-- Convert to a GMP_REAL_REF.
		do
			create Result.make_from_gmp_rational (Current)
		end

	to_hex_string: STRING is
			-- Convert to a hexadecimal string
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (numerator.item, feature {GMP_BASE_CONSTANTS}.Hexadecimal_base) +
			                     mpz_sizeinbase (denominator.item, feature {GMP_BASE_CONSTANTS}.Hexadecimal_base) +
			                     String_extra_bytes)
			mpq_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Hexadecimal_base, item)
			Result := s.string
		end

	to_mpfr: MPFR_REF is
			-- Convert to a MPFR_REF.
		do
			Result := to_gmp_real.to_mpfr
		end

	to_octal_string: STRING is
			-- Convert to a hexadecimal string
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (numerator.item, feature {GMP_BASE_CONSTANTS}.Octal_base) +
			                     mpz_sizeinbase (denominator.item, feature {GMP_BASE_CONSTANTS}.Octal_base) +
			                     String_extra_bytes)
			mpq_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Octal_base, item)
			Result := s.string
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
		do
			mpq_set (item, other.item)
		end

feature -- Basic operations

	abs: like Current is
			-- Absolute value.
		do
			create Result
			mpq_abs (Result.item, item)
		end

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		do
			create Result
			mpq_add (Result.item, item, other.item)
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'
		do
			create Result
			mpq_sub (Result.item, item, other.item)
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'
		do
			create Result
			mpq_mul (Result.item, item, other.item)
		end

	infix "/" (other: like Current): like Current is
			-- Division by `other'
		do
			create Result
			mpq_div (Result.item, item, other.item)
		end

	infix "^" (other: NUMERIC): like Current is
			-- Raise to the power `other'
		require
			good_exponent: exponentiable (other)
		local
			int_8_ref: INTEGER_8_REF
			int_16_ref: INTEGER_16_REF
			int_ref: INTEGER_REF
		do
			int_ref ?= other
			if int_ref /= Void then
				if int_ref.item >= 0 then
					create Result.make (numerator ^ int_ref.item, denominator ^ int_ref.item)
				else
					create Result.make (denominator ^ int_ref.item.abs, numerator ^ int_ref.item.abs)
				end
			else
				int_16_ref ?= other
				if int_16_ref /= Void then
					if int_16_ref.item >= 0 then
						create Result.make (numerator ^ int_16_ref.item, denominator ^ int_16_ref.item)
					else
						create Result.make (denominator ^ int_16_ref.item.abs, numerator ^ int_16_ref.item.abs)
					end
				else
					int_8_ref ?= other
					if int_8_ref /= Void then
						if int_8_ref.item >= 0 then
							create Result.make (numerator ^ int_8_ref.item, denominator ^ int_8_ref.item)
						else
							create Result.make (denominator ^ int_8_ref.item.abs, numerator ^ int_8_ref.item.abs)
						end
					end
				end
			end
		ensure
			result_exists: Result /= Void
		end

	prefix "+": like Current is
			-- Unary plus
		do
			create Result.make_from_gmp_rational (Current)
		end

	prefix "-": like Current is
			-- Unary minus
		do
			create Result
			mpq_neg (Result.item, item)
		end

feature -- Output

	out: STRING is
			-- Simple string representation.
		local
			s: C_STRING
		do
			create s.make_empty (mpz_sizeinbase (numerator.item, feature {GMP_BASE_CONSTANTS}.Binary_base) +
			                     mpz_sizeinbase (denominator.item, feature {GMP_BASE_CONSTANTS}.Binary_base) +
			                     String_extra_bytes)
			mpq_get_str (s.item, feature {GMP_BASE_CONSTANTS}.Decimal_base, item)
			Result := s.string
		end

feature {NONE} -- Implementation

	String_extra_bytes: INTEGER is 3

feature {NONE} -- Externals

	c_structure_size: INTEGER is
			-- Size of  "C" struct mpz_t.
		external
			"C [macro %"gmp.h%"]"
		alias
			"sizeof (__mpq_struct)"
		end

end -- class GMP_RATIONAL_REF
