indexing
	description: "Complex numbers with real coordinates of arbitrary precision"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	GMP_REAL_COMPLEX_REF

inherit
	COMPLEX [GMP_REAL_REF]
		redefine
			default_create
		end

create
	default_create, make, make_from_mpfr_complex, make_from_real, make_from_real_reference

feature {NONE} -- Initialization

	default_create is
			-- Initialize `Current'.
		do
			create real
			create imaginary
		end

	make (re, im: DOUBLE) is
			-- Set `real' from `re' and `imaginary' from `im'
		do
			create real.make (re)
			create imaginary.make (im)
		end

	make_from_mpfr_complex (c: MPFR_COMPLEX_REF) is
			-- Set value from `c'
		do
			create real.make_from_gmp_mpfr (c.real)
			create imaginary.make_from_gmp_mpfr (c.imaginary)
			normalize_precision
		end

	make_from_real (re, im: GMP_REAL_REF) is
			-- Set `real' from `re' and `imaginary' from `im'
		require
			re_exists: re /= Void
			im_exists: im /= Void
		do
			create real.make_from_gmp_real (re)
			create imaginary.make_from_gmp_real (im)
			normalize_precision
		ensure
			real_set: real.is_equal (re)
			imaginary_set: imaginary.is_equal (im)
		end

	make_from_real_reference (re, im: GMP_REAL_REF) is
			-- Set `real' from `re' and `imaginary' from `im'
		require
			re_exists: re /= Void
			im_exists: im /= Void
		do
			real := re
			imaginary := im
			normalize_precision
		ensure
			real_set: real = re
			imaginary_set: imaginary = im
		end

feature -- Access

	conjugate: like Current is
			-- Conjugate
		do
			create Result.make_from_real (real, -imaginary)
		end

	i: like Current is
			-- Square root of -1
		do
			create Result.make_from_real_reference (real.zero, imaginary.one)
		end

	one: like Current is
			-- Neutral element for "*" and "/"
		do
			create Result.make_from_real_reference (real.one, imaginary.zero)
		end

	zero: like Current is
			-- Neutral element for "+" and "-"
		do
			create Result.make_from_real_reference (real.zero, imaginary.zero)
		end

feature -- Status report

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		local
			int_ref: INTEGER_REF
			double_ref: DOUBLE_REF
			complex_ref: like Current
		do
			complex_ref ?= other
			if complex_ref /= Void then
				Result := (not is_negative_real (complex_ref))
				           or else
				          (not (real.is_equal (real.zero) and then imaginary.is_equal (imaginary.zero)))
			else
				double_ref ?= other
				if double_ref /= Void then
					Result := (double_ref.item > 0.0)
					           or else
					          (not (real.is_equal (real.zero) and then imaginary.is_equal (imaginary.zero)))
				else
					int_ref ?= other
					if int_ref /= Void then
						Result := (int_ref.item > 0)
						           or else
						          (not (real.is_equal (real.zero) and then imaginary.is_equal (imaginary.zero)))
					end
				end
			end
		end

feature -- Element change

	set (re, im: like real) is
			-- Set `real' to `re' and `imaginary' to `im'
		do
			real.copy (re)
			imaginary.copy (im)
			normalize_precision
		end

	set_imaginary (im: like real) is
			-- Set `imaginary' to `im'
		do
			imaginary.copy (im)
			normalize_precision
		end

	set_real (re: like real) is
			-- Set `real' to `re'
		do
			real.copy (re)
			normalize_precision
		end

feature -- Basic operations

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		do
			create Result.make_from_real_reference (real + other.real, imaginary + other.imaginary)
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'
		do
			create Result.make_from_real_reference (real - other.real, imaginary - other.imaginary)
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'
		do
			create Result.make_from_real_reference ((real * other.real) - (imaginary * other.imaginary),
			                                        (real * other.imaginary) + (other.real * imaginary))
		end

	infix "/" (other: like Current): like Current is
			-- Division by `other'
		local
			den: like real
		do
			den := other.real * other.real + other.imaginary * other.imaginary
			create Result.make_from_real_reference ((real * other.real + imaginary * other.imaginary) / den,
			                                        (imaginary * other.real - real * other.imaginary) / den)
		end

	infix "^" (other: NUMERIC): like Current is
			-- Raise to the power `other'
		local
			int_ref: INTEGER_REF
			double_ref: DOUBLE_REF
			complex_ref: like Current
			f: MPFR_COMPLEX_MATH
			res: MPFR_COMPLEX_REF
		do
			create f
			complex_ref ?= other
			if complex_ref /= Void then
				res := f.power (0,
				                create {MPFR_COMPLEX_REF}.make_from_real_complex (Current),
				                create {MPFR_COMPLEX_REF}.make_from_real_complex (complex_ref))
			else
				double_ref ?= other
				if double_ref /= Void then
					res := f.power (0,
					                create {MPFR_COMPLEX_REF}.make_from_real_complex (Current),
					                create {MPFR_COMPLEX_REF}.make (double_ref.item, 0.0))
				else
					int_ref ?= other
					if int_ref /= Void then
						res := f.power (0,
						                create {MPFR_COMPLEX_REF}.make_from_real_complex (Current),
						                create {MPFR_COMPLEX_REF}.make_from_integer (int_ref.item, 0))
					end
				end
			end
			create Result.make_from_mpfr_complex (res)
		end

	prefix "+": like Current is
			-- Unary plus
		do
			create Result.make_from_real (real, imaginary)
		end

	prefix "-": like Current is
			-- Unary minus
		do
			create Result.make_from_real_reference (-real, -imaginary)
		end

feature {NONE} -- Implementation

	is_negative_real (other: like Current): BOOLEAN is
			-- Is `other' a negative real number?
		do
			Result := (other.real < other.real.zero) and then other.imaginary.is_equal (imaginary.zero)
		end

	normalize_precision is
			-- Set the precision of `real' and `imaginary' to
			-- the maximum precision of `real' and `imaginary'
		local
			tmp: INTEGER
		do
			tmp := real.precision.max (imaginary.precision)
			real.set_precision (tmp)
			imaginary.set_precision (tmp)
		end

	Real_one: MPFR_REF is
		once
			create Result.make (1.0)
		end

end -- class GMP_REAL_COMPLEX_REF
