indexing
	description: "Real numbers with arbitrary precision"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	GMP_REAL_REF

inherit
	COMPARABLE
		redefine
			copy, default_create, infix ">", is_equal, out
		end

	MEMORY
		export
			{NONE} all
		redefine
			copy, default_create, dispose, is_equal, out
		end

	MEMORY_STRUCTURE
		rename
			make as memory_structure_make
		export
			{NONE} make_by_pointer, memory_structure_make
		redefine
			copy, default_create, is_equal, out
		end

	NUMERIC
		redefine
			copy, default_create, is_equal, out
		end

	GMP_REAL_FUNCTIONS
		export
			{NONE} all
		redefine
			copy, default_create, is_equal, out
		end

create
	default_create, make, make_from_gmp_integer, make_from_gmp_mpfr,
	make_from_gmp_rational, make_from_gmp_real, make_from_integer,
	make_from_string, make_precision

feature {NONE} -- Initialization

	default_create is
			-- Initialize `Current'.
		do
			memory_structure_make
			mpf_init (item)
		end

	make (d: DOUBLE) is
			-- Initialize from `d'.
		do
			memory_structure_make
			mpf_init_set_d (item, d)
		end

	make_from_gmp_integer (i: GMP_INTEGER_REF) is
			-- Initialize from `i'.
		require
			i_exists: i /= Void
		do
			memory_structure_make
			mpf_init (item)
			set_precision (i.precision + 1) --`i.precision' may be 0
			mpf_set_z (item, i.item)
		end
		
	make_from_gmp_mpfr (r: MPFR_REF) is
			-- Initialize from `r'.
		require
			r_exists: r /= Void
		local
			a: ANY
		do
			memory_structure_make
			mpf_init2 (item, r.precision)
			a := r.out.to_c
			mpf_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Decimal_base)
		end

	make_from_gmp_rational (r: GMP_RATIONAL_REF) is
			-- Initialize from `r'.
		require
			r_exists: r /= Void
		do
			memory_structure_make
			mpf_init2 (item, r.numerator.precision. max (r.denominator.precision))
			mpf_set_q (item, r.item)
		end
		
	make_from_gmp_real (r: GMP_REAL_REF) is
			-- Initialize from `r'.
		require
			r_exists: r /= Void
		do
			memory_structure_make
			mpf_init2 (item, r.precision)
			mpf_set (item, r.item)
		end

	make_from_integer (i: INTEGER) is
			-- Initialize from `i'.
		do
			memory_structure_make
			mpf_init_set_si (item, i)
		end

	make_from_string (s: STRING) is
			-- Initialize from `s'.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
		do
			memory_structure_make
			a := s.to_c
			mpf_init_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Decimal_base)
		end

	make_precision (p: INTEGER) is
			-- Initialize with `precision' set to `p'.
		do
			memory_structure_make
			mpf_init2 (item, p)
		end

feature -- Access

	one: like Current is
			-- Neutral element for "*" and "/"
		do
			create Result.make_precision (precision)
			Result.set_from_double (1.0)
		end

	zero: like Current is
			-- Neutral element for "+" and "-"
		do
			create Result.make_precision (precision)
		end

feature -- Measurement

	precision: INTEGER is
			-- Precision
		do
			Result := mpf_get_prec (item).abs
		end

	structure_size: INTEGER is
			-- Size to allocate (in bytes)
		do
			Result := c_structure_size
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
		do
			Result := not other.is_equal (zero)
		end

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		local
			int_ref: INTEGER_REF
		do
			int_ref ?= other
			Result := int_ref /= Void
			if int_ref /= Void then
				Result := (int_ref.item >= 0) or else (not is_equal (zero))
			end
		end

	fits_integer: BOOLEAN is
			-- Can `Current' be converted to an INTEGER without error?
		do
			Result := mpf_fits_slong_p (item) /= 0
		end

	is_integer: BOOLEAN is
			-- Is `Current' an integer?
		do
			Result := mpf_integer_p (item) /= 0
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN is
			-- Is current object less than `other'?
		do
			Result := mpf_cmp (item, other.item) < 0
		end

	infix ">" (other: like Current): BOOLEAN is
			-- Is current object greater than `other'?
		do
			Result := mpf_cmp (item, other.item) > 0
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := mpf_cmp (item, other.item) = 0
		end

feature -- Element change

	set_precision (a_precision: INTEGER) is
			-- Set `precision' to `a_precision'.
		require
			a_precision_positive: a_precision > 0
		do
			mpf_set_prec (item, a_precision)
		end

	set_from_double (d: DOUBLE) is
			-- Set from `d'.
		do
			mpf_set_d (item, d)
		end

	set_from_gmp_integer (i: GMP_INTEGER_REF) is
			-- Set from `i'.
		require
			i_exists: i /= Void
		do
			set_precision (i.precision)
			mpf_set_z (item, i.item)
		end

	set_from_gmp_real (r: GMP_REAL_REF) is
			-- Set from `r'.
		require
			r_exists: r /= Void
		do
			set_precision (r.precision)
			mpf_set (item, r.item)
		end

	set_from_integer (i: INTEGER) is
			-- Set from `i'.
		do
			mpf_set_si (item, i)
		end

	set_from_string (s: STRING) is
			-- Set from `s'.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
		do
			a := s.to_c
			mpf_set_str (item, $a, feature {GMP_BASE_CONSTANTS}.Decimal_base)
		end

feature -- Removal

	dispose is
			-- Action to be executed just before garbage collection
			-- reclaims an object.
		do
			mpf_clear (item)
		end

feature -- Conversion

	ceiling: GMP_INTEGER_REF is
			-- Smallest integral value no smaller than current object
		local
			temp: like Current
		do
			create temp
			temp.set_precision (precision)
			mpf_ceil (temp.item, item)
			create Result.make_from_gmp_real (temp)
		ensure
			result_no_smaller: Result.to_gmp_real >= Current
			close_enough: (Result.to_gmp_real - Current) < one
		end

	floor: GMP_INTEGER_REF is
			-- Greatest integral value no greater than current object
		local
			temp: like Current
		do
			create temp
			temp.set_precision (precision)
			mpf_floor (temp.item, item)
			create Result.make_from_gmp_real (temp)
		ensure
			result_no_greater: Result.to_gmp_real <= Current
			close_enough: (Current - Result.to_gmp_real) < one
		end

	to_double: DOUBLE is
			-- Convert to a DOUBLE.
		do
			Result := mpf_get_d (item)
		end

	to_gmp_integer: GMP_INTEGER_REF is
			-- Comvert to a GMP_INTEGER_REF.
		do
			create Result.make_from_gmp_real (Current)
		end

	to_mpfr: MPFR_REF is
			-- Convert to a MPFR_REF
		do
			create Result.make_from_gmp_real (Current)
		end

	to_gmp_rational: GMP_RATIONAL_REF is
			-- Convert to a GMP_RATIONAL_REF.
		do
			create Result.make_from_gmp_real (Current)
		end

	to_integer: INTEGER is
			-- Convert to an integer.
		require
			fits_integer: fits_integer
		do
			Result := mpf_get_si (item)
		end

	truncated: GMP_INTEGER_REF is
			-- Integer part (Same sign, largest absolute
			-- value no greater than current object's)
		local
			temp: like Current
		do
			create temp
			temp.set_precision (precision)
			mpf_trunc (temp.item, item)
			create Result.make_from_gmp_real (temp)
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
		do
			set_precision (other.precision)
			mpf_set (item, other.item)
		ensure then
			same_precision: precision = other.precision
		end

feature -- Basic operations

	abs: like Current is
			-- Absolute value.
		do
			create Result.make_precision (precision)
			mpf_abs (Result.item, item)
		end

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		do
			create Result.make_precision (precision.max (other.precision))
			mpf_add (Result.item, item, other.item)
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'
		do
			create Result.make_precision (precision.max (other.precision))
			mpf_sub (Result.item, item, other.item)
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'
		do
			create Result.make_precision (precision.max (other.precision))
			mpf_mul (Result.item, item, other.item)
		end

	infix "/" (other: like Current): like Current is
			-- Division by `other'
		do
			create Result.make_precision (precision.max (other.precision))
			mpf_div (Result.item, item, other.item)
		end

	infix "^" (i: INTEGER): like Current is
			-- Raise to the power `i'.
		require
			good_exponent: exponentiable (i)
		do
			create Result.make_precision (precision)
			if i < 0 then
				mpf_pow_ui (Result.item, (one / Current).item, i.abs)
			else
				mpf_pow_ui (Result.item, item, i)
			end
		ensure
			result_exists: Result /= Void
		end

	prefix "+": like Current is
			-- Unary plus
		do
			create Result.make_from_gmp_real (Current)
		end

	prefix "-": like Current is
			-- Unary minus
		do
			create Result.make_precision (precision)
			mpf_neg (Result.item, item) 
		end

feature -- Output

	out: STRING is
			-- Decimal string representation.
		local
			sgn, temp: STRING
			exp, i: INTEGER
			p: POINTER
		do
			if Current < zero then
				sgn := "-"
			else
				sgn := ""
			end
			p := mpf_get_str (default_pointer,
			                  $exp,
			                  feature {GMP_BASE_CONSTANTS}.Decimal_base,
			                  0,
			                  item)
			temp := (create {C_STRING}.make_by_pointer (p)).string
			c_free (p)
			temp.prune ('-')
			if temp.is_empty then
				Result := "0"
			else
				if exp > 0 then
					if exp <= temp.count then
						Result := temp.substring (1, exp)
					end
					if temp.count > exp then
						Result.append_character ('.')
						Result.append_string (temp.substring (exp + 1, temp.count))
					elseif temp.count < exp then
						from
							Result := temp.substring (1, temp.count)
						until
							Result.count = exp
						loop
							Result.append_character ('0')
						end
					end
				elseif exp <= 0 then
					Result := "0."
					from
						i := -1
					until
						i < exp
					loop
						Result.append_character ('0')
						i := i - 1
					end
					Result.append_string (temp)
				end
			end
			Result.prepend_string (sgn)
		end

feature {NONE} -- Externals

	c_free (mem: POINTER) is
		external
			"C (void *) | <malloc.h>"
		alias
			"free"
		end

	c_structure_size: INTEGER is
			-- Size of  "C" struct mpz_t.
		external
			"C [macro %"gmp.h%"]"
		alias
			"sizeof (__mpf_struct)"
		end

end -- class GMP_REAL_REF
