indexing
	description: "Complex numbers with real coordinates of arbitrary precision and adjustable rounding mode"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	MPFR_COMPLEX_REF

inherit
	COMPLEX [MPFR_REF]
		redefine
			default_create, divisible
		end

	MPFR_CONSTANTS
		undefine
			copy, is_equal, out
		redefine
			default_create
		end

	MPFR_FUNCTIONS
		export
			{NONE} all
		undefine
			copy, is_equal, out
		redefine
			default_create
		end

	MPFR_MATH
		export
			{NONE} all
		undefine
			copy, is_equal, out
		redefine
			default_create
		end

create
	default_create, make, make_from_integer, make_from_mpfr,
	make_from_mpfr_imaginary, make_from_mpfr_real,
	make_from_mpfr_reference, make_from_real_complex

feature {NONE} -- Initialization

	default_create is
			-- Initialize `Current'.
		do
			create real
			create imaginary
		end

	make (re, im: DOUBLE) is
			-- Set `real' from `re' and `imaginary' from `im'
		do
			create real.make (re)
			create imaginary.make (im)
			normalize_precision
		end

	make_from_integer (re, im: INTEGER) is
			-- Set `real' from `re' and `imaginary' from `im'
		do
			create real.make_from_integer (re)
			create imaginary.make_from_integer (im)
		end

	make_from_mpfr (re, im: MPFR_REF) is
			-- Set `real' from `re' and `imaginary' from `im'
		require
			re_exists: re /= Void
			im_exists: im /= Void
		do
			create real.make_from_mpfr (re)
			create imaginary.make_from_mpfr (im)
			normalize_precision
		ensure
			real_set: real.is_equal (re)
			imaginary_set: imaginary.is_equal (im)
		end

	make_from_mpfr_imaginary (im: MPFR_REF) is
			-- Set `imaginary' from `im'
		require
			im_exists: im /= Void
		do
			create real
			real.set_rounding_mode (im.rounding_mode)
			create imaginary.make_from_mpfr (im)
			normalize_precision
		ensure
			real_set: real.is_equal (real.zero)
			imaginary_set: imaginary.is_equal (im)
		end

	make_from_mpfr_real (re: MPFR_REF) is
			-- Set `real' from `re'
		require
			re_exists: re /= Void
		do
			create real.make_from_mpfr (re)
			create imaginary
			imaginary.set_rounding_mode (re.rounding_mode)
			normalize_precision
		ensure
			real_set: real.is_equal (re)
			imaginary_set: imaginary.is_equal (imaginary.zero)
		end

	make_from_mpfr_reference (re, im: MPFR_REF) is
			-- Set `real' from `re' and `imaginary' from `im'
		require
			re_exists: re /= Void
			im_exists: im /= Void
		do
			real := re
			imaginary := im
			normalize_precision
		ensure
			real_set: real = re
			imaginary_set: imaginary = im
		end

	make_from_real_complex (c: GMP_REAL_COMPLEX_REF) is
			-- 
		do
			create real.make_from_gmp_real (c.real)
			create imaginary.make_from_gmp_real (c.imaginary)
		end

feature -- Access

	conjugate: like Current is
			-- Conjugate
		do
			create Result.make_from_mpfr (real, -imaginary)
		end

	i: like Current is
			-- Square root of -1
		do
			create Result.make_from_mpfr_reference (real.zero, imaginary.one)
		end

--	imaginary: like real
			-- Imaginary part

	one: like Current is
			-- Neutral element for "*" and "/"
		do
			create Result.make_from_mpfr_reference (real.one, imaginary.zero)
		end

--	real: MPFR_REF
			-- Real part

	zero: like Current is
			-- Neutral element for "+" and "-"
		do
			create Result.make_from_mpfr_reference (real.zero, imaginary.zero)
		end

feature -- Measurement

	precision: INTEGER is
			-- Precision
		do
			Result := real.precision.max (imaginary.precision)
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
		do
			Result := True
		end

	exponentiable (other: NUMERIC): BOOLEAN is
			-- May current object be elevated to the power `other'?
		local
			int_ref: INTEGER_REF
			double_ref: DOUBLE_REF
			complex_ref: like Current
		do
			complex_ref ?= other
			if complex_ref /= Void then
				Result := (not is_negative_real (complex_ref))
				           or else
				          (not (real.is_equal (real.zero) and then imaginary.is_equal (imaginary.zero)))
			else
				double_ref ?= other
				if double_ref /= Void then
					Result := (double_ref.item > 0.0)
					           or else
					          (not (real.is_equal (real.zero) and then imaginary.is_equal (imaginary.zero)))
				else
					int_ref ?= other
					if int_ref /= Void then
						Result := (int_ref.item > 0)
						           or else
						          (not (real.is_equal (real.zero) and then imaginary.is_equal (imaginary.zero)))
					end
				end
			end
		end

	is_infinite: BOOLEAN is
			-- Is either `real' or `imaginary' infinite?
		do
			Result := real.is_infinite or else imaginary.is_infinite
		end

	is_not_a_number: BOOLEAN is
			-- Is either `real' or `imaginary' not a number?
		do
			Result := real.is_not_a_number or else imaginary.is_not_a_number
		end

	rounding_mode: INTEGER is
			-- Rounding mode
		do
			Result := real.rounding_mode
		end

feature -- Status setting

	set_rounding_mode (a_mode: INTEGER) is
			-- Set `rounding_mode' to `a_mode'
		require
			valid_mode: (a_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn) or else
			            (a_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndz) or else
			            (a_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndu) or else
			            (a_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndd)
		do
			real.set_rounding_mode (a_mode)
			imaginary.set_rounding_mode (a_mode)
		ensure
			rounding_mode_set: rounding_mode = a_mode
		end

feature -- Element change

	set (re, im: like real) is
			-- Set `real' to `re' and `imaginary' to `im'
		do
			real.copy (re)
			imaginary.copy (im)
			normalize_precision
		end

	set_from_double (re, im: DOUBLE) is
			-- Set `real' to `re' and `imaginary' to `im'
		do
			real.set_from_double (re)
			imaginary.set_from_double (im)
		end

	set_imaginary (im: like real) is
			-- Set `imaginary' to `im'
		do
			imaginary.copy (im)
			normalize_precision
		end

	set_imaginary_from_double (im: DOUBLE) is
			-- Set `imaginary' to `im'
		do
			imaginary.set_from_double (im)
		end

	set_precision (prec: INTEGER) is
			-- Set `precision' to `prec'
		do
			real.set_precision (prec)
			imaginary.set_precision (prec)
		end

	set_real (re: like real) is
			-- Set `real' to `re'
		do
			real.copy (re)
			normalize_precision
		end

	set_real_from_double (re, im: DOUBLE) is
			-- Set `real' to `re'
		do
			real.set_from_double (re)
		end

feature -- Basic operations

	abs: like real is
			-- Absolute value
		do
			Result := hypotenuse (real, imaginary)
		ensure
			not_negative: not (Result < real.zero)
		end

	add_real (r: like real): like Current is
			-- Add real number `r' to `Current'
		require
			r_exists: r /= Void
		do
			create Result.make_from_mpfr (real + r, imaginary)
		end

	arg: like real is
			-- Counter-clockwise angle from positive x-axis in radians
		require
			not_zero: not (real.is_equal (real.zero) and imaginary.is_equal (imaginary.zero))
		do
			if ((real >= Mpfr_zero) and then (imaginary >= Mpfr_zero)) then
				Result := arc_tangent (imaginary / real)
			elseif ((real < Mpfr_zero) and then (imaginary >= Mpfr_zero)) then
				Result := arc_tangent (imaginary / real) + pi (precision)
			elseif ((real <= Mpfr_zero) and then (imaginary < Mpfr_zero)) then
				Result := arc_tangent (imaginary / real) - pi (precision)
			elseif ((real >= Mpfr_zero) and then (imaginary < Mpfr_zero)) then
				Result := arc_tangent (imaginary / real)
			end
		end

	imaginary_multiply (an_imaginary: MPFR_REF): like Current is
			-- Mulitply by imaginary number `an_imaginary'
		do
			create Result.make_from_mpfr_reference (-imaginary * an_imaginary, real * an_imaginary)
		end

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		do
			create Result.make_from_mpfr_reference (real + other.real, imaginary + other.imaginary)
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'
		do
			create Result.make_from_mpfr_reference (real - other.real, imaginary - other.imaginary)
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'
		do
			create Result.make_from_mpfr_reference ((real * other.real) - (imaginary * other.imaginary),
			                                        (real * other.imaginary) + (other.real * imaginary))
		end

	infix "/" (other: like Current): like Current is
			-- Division by `other'
		local
			s, sbr, sbi: like real
		do
			s := Mpfr_one / other.abs
			sbr := s * other.real
			sbi := s * other.imaginary
			create Result.make_from_mpfr_reference ((real * sbr + imaginary * sbi) * s,
			                                        (imaginary * sbr - real * sbi) * s)
		end

	infix "^" (other: NUMERIC): like Current is
			-- Raise to the power `other'
		local
			int_ref: INTEGER_REF
			double_ref: DOUBLE_REF
			complex_ref: like Current
			f: MPFR_COMPLEX_MATH
		do
			create f
			complex_ref ?= other
			if complex_ref /= Void then
				Result := f.power (0, Current, complex_ref)
			else
				double_ref ?= other
				if double_ref /= Void then
					Result := f.power (0, Current, create {MPFR_COMPLEX_REF}.make (double_ref.item, 0.0))
				else
					int_ref ?= other
					if int_ref /= Void then
						Result := f.power (0, Current, create {MPFR_COMPLEX_REF}.make_from_integer (int_ref.item, 0))
					end
				end
			end
		end

	multiply_i: like Current is
			-- `Current' multiplied by `i'
		do
			create Result.make_from_mpfr (-imaginary, real)
		end

	prefix "+": like Current is
			-- Unary plus
		do
			create Result.make_from_mpfr (real, imaginary)
		end

	prefix "-": like Current is
			-- Unary minus
		do
			create Result.make_from_mpfr_reference (-real, -imaginary)
		end

	reciprocal: like Current is
			-- 1 / `Current'
		local
			div: MPFR_REF
		do
			div := real * real + imaginary * imaginary
			create Result.make_from_mpfr_reference (real / div, -imaginary / div)
		end

	scale, real_multiply (a_real: MPFR_REF): like Current is
			-- Multiply by real number `a_real'
		do
			create Result.make_from_mpfr_reference (real * a_real, imaginary * a_real)
		end

	square: MPFR_COMPLEX_REF is
			-- Raise to the power 2
		do
			create Result.make_from_mpfr_reference ((real * real) - (imaginary * imaginary),
			                                        Mpfr_two * real * imaginary)
		end

feature {NONE} -- Implementation

	is_negative_real (other: like Current): BOOLEAN is
			-- Is `other' a negative real number?
		do
			Result := (other.real < other.real.zero) and then other.imaginary.is_equal (imaginary.zero)
		end

	normalize_precision is
			-- Set the precision of `real' and `imaginary' to
			-- the maximum precision of `real' and `imaginary'
		local
			tmp: INTEGER
		do
			tmp := real.precision.max (imaginary.precision)
			real.set_precision (tmp)
			imaginary.set_precision (tmp)
		end

invariant
	same_rounding_modes: imaginary.rounding_mode = real.rounding_mode

end -- class MPFR_COMPLEX_REF
