indexing
	description: "Real numbers with arbitrary precision and adjustable rounding mode"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	MPFR_REF

inherit
	COMPARABLE
		redefine
			copy, default_create, out
		end

	MPFR_CONSTANTS
		undefine
			is_equal
		redefine
			copy, default_create, out
		end

	MPFR_FUNCTIONS
		export
			{NONE} all
		undefine
			is_equal
		redefine
			copy, default_create, out
		end

	MEMORY
		export
			{NONE} all
		undefine
			is_equal
		redefine
			copy, default_create, dispose, out
		end

	MEMORY_STRUCTURE
		rename
			make as memory_structure_make
		export
			{NONE} make_by_pointer, memory_structure_make
		undefine
			is_equal
		redefine
			copy, default_create, out
		end

	NUMERIC
		undefine
			is_equal
		redefine
			copy, default_create, out
		end

create
	default_create, make, make_by_pointer, make_from_gmp_real,
	make_from_integer, make_from_mpfr, make_from_string

feature {NONE} -- Initialization

	default_create is
			-- 
		local
			res: INTEGER
		do
			memory_structure_make
			rounding_mode := feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn
			res := mpfr_init_set_d (item, 0.0, rounding_mode)
		end

	make (d: DOUBLE) is
			-- Initialize `Current' from `d'.
		local
			res: INTEGER
		do
			memory_structure_make
			rounding_mode := feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn
			res := mpfr_init_set_d (item, d, rounding_mode)
		end

	make_from_gmp_real (r: GMP_REAL_REF) is
			-- Initialize form `r'
		require
			r_exists: r /= Void
		local
			res: INTEGER
		do
			default_create
			set_precision (r.precision)
			res := mpfr_set_f (item, r.item, rounding_mode)
		end

	make_from_integer (i: INTEGER) is
			-- Initialize `Current' from `i'.
		local
			res: INTEGER
		do
			memory_structure_make
			rounding_mode := feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn
			res := mpfr_init_set_si (item, i, rounding_mode)
		end

	make_from_mpfr (r: MPFR_REF) is
			-- Initialize form `r'
		require
			r_exists: r /= Void
		local
			res: INTEGER
		do
			default_create
			set_precision (r.precision)
			rounding_mode := r.rounding_mode
			res := mpfr_set (item, r.item, rounding_mode)
		end

	make_from_string (s: STRING) is
			-- Initialize from `s'.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
			res: INTEGER
		do
			memory_structure_make
			rounding_mode := feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn
			a := s.to_c
			res := mpfr_init_set_str (item,
			                          $a,
			                          feature {GMP_BASE_CONSTANTS}.Decimal_base,
			                          rounding_mode)
		end

feature -- Access

	negative_infinity: like Current is
			-- Negative infinity
		do
			create Result
			mpfr_set_inf (Result.item, -1)
		end

	not_a_number: like Current is
			-- Nan
		do
			create Result
			mpfr_set_nan (Result.item)
		end

	one: like Current is
			-- Neutral element for "*" and "/"
		do
			create Result
			Result.set_precision (precision)
			Result.set_from_double (1.0)
		end

	positive_infinity: like Current is
			-- Positive infinity
		do
			create Result
			mpfr_set_inf (Result.item, 1)
		end

	sign: INTEGER is
			-- 1 if positive
			-- 0 if `zero'
			-- -1 if negative
		do
			Result := mpfr_sgn (item)
		end

	zero: like Current is
			-- Neutral element for "+" and "-"
		do
			create Result
			Result.set_precision (precision)
		end

feature -- Measurement

	precision: INTEGER is
			-- Number of bits used in the mantissa.
		do
			Result := mpfr_get_prec (item)
		end

	structure_size: INTEGER is
			-- Size to allocate (in bytes)
		do
			Result := c_structure_size
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN is
			-- Is current object less than `other'?
		do
			Result := mpfr_cmp (item, other.item) < 0
		end

feature -- Status report

	divisible (other: like Current): BOOLEAN is
			-- May current object be divided by `other'?
		do
			Result := True
		end

	exponentiable (other: like Current): BOOLEAN is
			-- May current object be elevated to the power `other'?
		do
			Result := True
		end

	is_infinite: BOOLEAN is
			-- Is `Current' infinite?
		do
			Result := mpfr_inf_p (item) /= 0
		end

	is_not_a_number: BOOLEAN is
			-- Is `Current' Nan?
		do
			Result := mpfr_nan_p (item) /= 0
		end

	last_result_exact: BOOLEAN is
			-- Did the last mpfr function give an exact result?
		do
			Result := last_result = 0
		end

	last_result_greater: BOOLEAN is
			-- Did the last mpfr function give a result
			-- greater than the exact result?
		do
			Result := last_result > 0
		end

	last_result_less: BOOLEAN is
			-- Did the last mpfr function give a result
			-- less than the exact result?
		do
			Result := last_result < 0
		end

	rounding_mode: INTEGER
			-- Rounding mode

	is_valid_rounding_mode (a_rounding_mode: INTEGER): BOOLEAN is
			-- Is `' a valid rounding mode?
		do
			Result := (a_rounding_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn) or else
			          (a_rounding_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndz) or else
			          (a_rounding_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndu) or else
			          (a_rounding_mode = feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndd)
		end

feature -- Status setting

	set_rounding_mode (a_mode: INTEGER) is
			-- Set `rounding_mode' to `a_mode'
		require
			valid_rounding_mode: is_valid_rounding_mode (a_mode)
		do
			rounding_mode := a_mode
		ensure
			rounding_mode_set: rounding_mode = a_mode
		end

feature -- Element change

	set_from_double (d: DOUBLE) is
			-- Set from `d'.
		do
			last_result := mpfr_set_d (item, d, rounding_mode)
		end

	set_from_integer (i: INTEGER) is
			-- Set `Current' from `i'.
		do
			last_result := mpfr_init_set_si (item, i, rounding_mode)
		end

	set_from_string (s: STRING) is
			-- Set from `s'.
		require
			s_exists: s /= Void
			s_not_empty: not s.is_empty
		local
			a: ANY
		do
			a := s.to_c
			last_result := mpfr_set_str (item,
			                     $a,
			                     feature {GMP_BASE_CONSTANTS}.Decimal_base,
			                     rounding_mode)
		end

	set_precision (a_precision: INTEGER) is
			-- Set `precision' to `a_precision'.
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision) and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			last_result := mpfr_round_prec (item, rounding_mode, a_precision)
		end

feature -- Removal

	dispose is
			-- Action to be executed just before garbage collection
			-- reclaims an object.
		do
			mpfr_clear (item)
		end

feature -- Conversion

	ceiling: like Current is
			-- Smallest integral value no smaller than current object
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_ceil (Result.item, item)
		ensure
			result_no_smaller: Result >= Current
			close_enough: Result - Current < one
		end

	floor: like Current is
			-- Greatest integral value no greater than current object
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_floor (Result.item, item)
		ensure
			result_no_greater: Result <= Current
			close_enough: Current - Result < one
		end

	rounded: like Current is
			-- Rounded integral value
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_round (Result.item, item)
		end

	to_double: DOUBLE is
			-- Convert to a DOUBLE
		do
			Result := mpfr_get_d (item, rounding_mode)
		end

	to_gmp_integer: GMP_INTEGER_REF is
			-- Comvert to a GMP_INTEGER_REF.
		do
			create Result.make_from_gmp_real (to_gmp_real)
		end

	to_gmp_rational: GMP_RATIONAL_REF is
			-- Convert to a GMP_RATIONAL_REF.
		do
			Result := to_gmp_real.to_gmp_rational
		end

	to_gmp_real: GMP_REAL_REF is
			-- Convert to a GMP_REAL_REF
		do
			create Result.make_from_gmp_mpfr (Current)
		end

	truncated: like Current is
			-- Integer part (Same sign, largest absolute
			-- value no greater than current object's)
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_trunc (Result.item, item)
		end

feature -- Duplication

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to `other', so as to yield equal objects.
		do
			set_precision (other.precision)
			last_result := mpfr_set (item, other.item, other.rounding_mode)
			rounding_mode := other.rounding_mode
		ensure then
			same_precision: precision = other.precision
			same_rounding_mode: rounding_mode = other.rounding_mode
		end

feature -- Basic operations

	abs: like Current is
			-- Absolute value.
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_abs (Result.item, item, rounding_mode)
		end

	infix "+" (other: like Current): like Current is
			-- Sum with `other' (commutative).
		do
			create Result
			Result.set_precision (precision.max (other.precision))
			last_result := mpfr_add (Result.item, item, other.item, rounding_mode)
		end

	add (other: like Current) is
			-- Add `other' to `Current'
		require
			other_exists: other /= Void
		do
			last_result := mpfr_add (item, item, other.item, rounding_mode)
		end

	infix "-" (other: like Current): like Current is
			-- Result of subtracting `other'
		do
			create Result
			Result.set_precision (precision.max (other.precision))
			last_result := mpfr_sub (Result.item, item, other.item, rounding_mode)
		end

	sub (other: like Current) is
			-- Subtract `other' from `Current'
		require
			other_exists: other /= Void
		do
			last_result := mpfr_sub (item, item, other.item, rounding_mode)
		end

	infix "*" (other: like Current): like Current is
			-- Product by `other'
		do
			create Result
			Result.set_precision (precision.max (other.precision))
			last_result := mpfr_mul (Result.item, item, other.item, rounding_mode)
		end

	mul (other: like Current) is
			-- Multiply `Current' by `other'
		require
			other_exists: other /= Void
		do
			last_result := mpfr_mul (item, item, other.item, rounding_mode)
		end

	infix "/" (other: like Current): like Current is
			-- Division by `other'
		do
			create Result
			Result.set_precision (precision.max (other.precision))
			last_result := mpfr_div (Result.item, item, other.item, rounding_mode)
		end

	div (other: like Current) is
			-- Divide `Current' by `other'
		require
			other_exists: other /= Void
			good_divisor: divisible (other)
		do
			last_result := mpfr_div (item, item, other.item, rounding_mode)
		end

	div_2 is
			-- Divide by 2.0
		do
			last_result := mpfr_div (item, item, Mpfr_two.item, rounding_mode)
		end

	infix "^" (other: like Current): like Current is
			-- Raise to the power `other'.
		require
			good_exponent: exponentiable (other)
		do
			create Result
			Result.set_precision (precision)
			if other < zero then
				last_result := mpfr_pow (Result.item, (Mpfr_one / Current).item, other.abs.item, rounding_mode)
			else
				last_result := mpfr_pow (Result.item, item, other.item, rounding_mode)
			end
		ensure
			result_exists: Result /= Void
		end

	power (other: like Current) is
			-- Raise `Current' to the power `other'
		require
			good_exponent: exponentiable (other)
		do
			if other < zero then
				last_result := mpfr_pow (item, (Mpfr_one / Current).item, other.abs.item, rounding_mode)
			else
				last_result := mpfr_pow (item, item, other.item, rounding_mode)
			end
		end

	prefix "+": like Current is
			-- Unary plus
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_set (Result.item, item, rounding_mode)
		end

	prefix "-": like Current is
			-- Unary minus
		do
			create Result
			Result.set_precision (precision)
			last_result := mpfr_neg (Result.item, item, rounding_mode)
		end

	negate is
			-- Negate `Current'
		do
			last_result := mpfr_neg (item, item, rounding_mode)
		end

feature -- Output

	out: STRING is
			-- Decimal string representation.
		local
			temp: STRING
			exp, i: INTEGER
			p: POINTER
		do
			p := mpfr_get_str (default_pointer,
			                   $exp,
			                   feature {GMP_BASE_CONSTANTS}.Decimal_base,
			                   0,
			                   item,
			                   rounding_mode)
			temp := (create {C_STRING}.make_by_pointer (p)).string
			c_free (p)
			temp.prune ('-')
			if is_equal (zero) then
				Result := "0"
			elseif is_infinite then
				Result := "Inf"
			elseif is_not_a_number then
				Result := "Nan"
			else
				if sign = -1 then
					Result := "-"
				else
					Result := ""
				end
				if exp > 0 then
					if exp <= temp.count then
						Result.append_string (temp.substring (1, exp))
					end
					if temp.count > exp then
						Result.append_character ('.')
						Result.append_string (temp.substring (exp + 1, temp.count))
					elseif temp.count < exp then
						from
							Result.append_string (temp.substring (1, temp.count))
						until
							Result.count = exp
						loop
							Result.append_character ('0')
						end
					end
				else
					Result.append_string ("0.")
					from
						i := -1
					until
						i < exp
					loop
						Result.append_character ('0')
						i := i - 1
					end
					Result.append_string (temp)
				end
			end
			if Result.has ('.') then
				Result.prune_all_trailing ('0')
				if Result.item (Result.count) = '.' then
					Result.append_character ('0')
				end
			end
		end

	out2: STRING is
			-- Simple string representation: [mantissa, exponent]
		local
			expnt: INTEGER
			mant: GMP_INTEGER_REF
		do
			create mant
			expnt := mpfr_get_z_exp (mant.item, item)
			Result := "["
			Result.append_string (mant.out)
			Result.append_string (", ")
			Result.append_integer (expnt)
			Result.append_character (']')
		end

	out3: STRING is
			-- Simple string representation.
		local
			expnt: INTEGER
			mant: GMP_INTEGER_REF
		do
			if is_not_a_number then
				Result := "Nan"
			elseif is_infinite then
				if sign = -1 then
					Result := "-Inf"
				else
					Result := "+Inf"
				end
			else
				create mant
				expnt := mpfr_get_z_exp (mant.item, item)
				Result := (mant.to_gmp_real * (Two_const ^ expnt)).out
				if sign = -1 then
					Result.prepend_character ('-')
				end
			end
		end

feature {NONE} -- Implementation

	last_result: INTEGER

	Two_const: GMP_REAL_REF is
		once
			create Result.make (2.0)
		end

feature {NONE} -- Externals

	c_free (mem: POINTER) is
		external
			"C (void *) | <malloc.h>"
		alias
			"free"
		end
		
	c_structure_size: INTEGER is
			-- Size of  "C" struct mpfr_t.
		external
			"C [macro %"mpfr.h%"]"
		alias
			"sizeof (__mpfr_struct)"
		end

	mpfr_prec (p: POINTER): INTEGER is
		external
			"C [struct %"mpfr.h%"] (__mpfr_struct): EIF_INTEGER"
		alias
			"_mpfr_prec"
		end

	mpfr_size (p: POINTER): INTEGER is
		external
			"C [struct %"mpfr.h%"] (__mpfr_struct): EIF_INTEGER"
		alias
			"_mpfr_size"
		end

	mpfr_exponent (p: POINTER): INTEGER is
		external
			"C [struct %"mpfr.h%"] (__mpfr_struct): EIF_INTEGER"
		alias
			"_mpfr_exp"
		end

	mpfr_d (p: POINTER): POINTER is
		external
			"C [struct %"mpfr.h%"] (__mpfr_struct): EIF_POINTER"
		alias
			"_mpfr_d"
		end
		
end -- class MPFR_REF
