indexing
	description: "Math operations on GMP_INTEGER_REFs"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	GMP_INTEGER_MATH

inherit
	GMP_INTEGER_FUNCTIONS
		export
			{NONE} all
		end

feature -- Access

	binomial_coefficient (n: GMP_INTEGER_REF; k: INTEGER): GMP_INTEGER_REF is
			-- Binomial coefficient n over k
		require
			n_exists: n /= Void
			k_not_negative: not (k < 0)
		do
			create Result
			mpz_bin_ui (Result.item, n.item, k)
		end

	factorial (i: INTEGER): GMP_INTEGER_REF is
			-- Factorial of `i'
		require
			i_not_negative: not (i < 0)
		do
			create Result
			mpz_fac_ui (Result.item, i)
		end

	gcd (i, j: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- Greatest common divisor of `i' and `j'
		require
			i_exists: i /= Void
			j_exists: j /= Void
		do
			create Result
			mpz_gcd (Result.item, i.item, j.item)
		end

	gcd_ext (a, b: GMP_INTEGER_REF): TUPLE [GMP_INTEGER_REF, GMP_INTEGER_REF, GMP_INTEGER_REF] is
			-- Set `Result.item (1)' to the greatest common divisor of `a' and `b',
			-- and in addition set `Result.item (2)' and `Result.item (3)' to coefficients satisfying
			-- `a' * `Result.item (2)'' + `b' * `Result.item (3)' = `Result.item (1)'.
			-- `Result.item (1)' is always positive, even if one or both of `a' and `b' are negative.
		require
			a_exists: a /= Void
			b_exists: b /= Void
		local
			g, s, t: GMP_INTEGER_REF
		do
			create g
			create s
			create t
			mpz_gcdext (g.item, s.item, t.item, a.item, b.item)
			create Result.make
			Result.put (g, 1)
			Result.put (s, 2)
			Result.put (t, 3)
		end

	invert (i, mod: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- Inverse of `i' modulo `mod'
		require
			i_exists: i /= Void
			mod_exists: mod /= Void
		do
			create Result
			inverse_exists := mpz_invert (Result.item, i.item, mod.item) /= 0
		end

	jacobi (a, b: GMP_INTEGER_REF): INTEGER is
			-- Jacobi symbol (`a' /` b')
		require
			a_exists: a /= Void
			b_exists: b /= Void
		do
			Result := mpz_jacobi (a.item, b.item)
		end

	kronecker (a, b: GMP_INTEGER_REF): INTEGER is
			-- Jacobi symbol (`a' / `b') with the Kronecker extension
			-- (`a' / 2) = (2 / `a') when `a' odd, or (`a' / 2) = 0
			-- when `a' even
		require
			a_exists: a /= Void
			b_exists: b /= Void
		do
			Result := mpz_kronecker (a.item, b.item)
		end

	lcm (i, j: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- Least common multiple of `i' and `j'
		require
			i_exists: i /= Void
			j_exists: j /= Void
		do
			create Result
			mpz_lcm (Result.item, i.item, j.item)
		end

	legendre (a, p: GMP_INTEGER_REF): INTEGER is
			-- Legendre symbol (`a' / `p')
		require
			a_exists: a /= Void
			p_exists: p /= Void
			p_positive: p > p.zero
			p_odd: is_odd (p)
			p_is_prime: is_prime (p, Prime_test_reps)
		do
			Result := mpz_legendre (a.item, p.item)
		end

	mul_2_exp (base: GMP_INTEGER_REF; exp: INTEGER): GMP_INTEGER_REF is
			-- `base' times 2 ^ `exp'
		require
			base_exists: base /= Void
			exp_not_negative: not (exp < 0)
		do
			create Result
			mpz_mul_2exp (Result.item, base.item, exp)
		end

	next_prime (i: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- Next prime number greater than `i'
		require
			i_exists: i /= Void
		do
			create Result
			mpz_nextprime (Result.item, i.item)
		end

	remove (i, f: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- Remove all occurrences of the factor `f' from `i'
		require
			i_exists: i /= Void
			f_exists: f /= Void
		do
			create Result
			factors_removed := mpz_remove (Result.item, i.item, f.item)
		end

	root (i: GMP_INTEGER_REF; n: INTEGER): GMP_INTEGER_REF is
			-- Truncated integer part of the `n''th root of `i'.
		require
			i_exists: i /= Void
			n_positive: n > 0
		do
			create Result
			last_root_exact := mpz_root (Result.item, i.item, n) /= 0
		end

	square_root (i: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- Truncated integer part of the square root of `i'.
		require
			i_exists: i /= Void
		do
			create Result
			mpz_sqrt (Result.item, i.item)
		end

	swqare_root_remainder (i: GMP_INTEGER_REF): GMP_INTEGER_REF is
			-- `i' - (`square_root (i)' * `square_root (i)')
		require
			i_exists: i /= Void
		do
			create Result
			mpz_sqrtrem ((create {GMP_INTEGER_REF}.default_create).item, Result.item, i.item)
		end

feature -- Status report

	factors_removed: INTEGER
			-- Number of factors removed in the last call to `remove'

	inverse_exists: BOOLEAN
			-- Was there a defined result for the last call to `invert'?

	is_composite (i: GMP_INTEGER_REF; reps: INTEGER): BOOLEAN is
			-- Is `i' definetly a composite number?
		require
			i_exists: i /= Void
			reps_positive: reps > 0
		do
			Result := mpz_probab_prime_p (i.item, reps) = 0
		end

	is_even (i: GMP_INTEGER_REF): BOOLEAN is
			-- Is `i' even?
		require
			i_exists: i /= Void
		do
			Result := mpz_even_p (i.item) /= 0
		end

	is_odd (i: GMP_INTEGER_REF): BOOLEAN is
			-- Is `i' odd?
		require
			i_exists: i /= Void
		do
			Result := mpz_odd_p (i.item) /= 0
		end

	is_perfect_power (i: GMP_INTEGER_REF): BOOLEAN is
			-- Does there exist integers a and b, with b > 1,
			-- such that `i' equals a raised to the power b?
		require
			i_exists: i /= Void
		do
			Result := mpz_perfect_power_p (i.item)  /= 0
		end

	is_perfect_square (i: GMP_INTEGER_REF): BOOLEAN is
			-- Is `i' a perfect square?
		require
			i_exists: i /= Void
		do
			Result := mpz_perfect_square_p (i.item) /= 0
		end

	is_prime (i: GMP_INTEGER_REF; reps: INTEGER): BOOLEAN is
			-- Is `i' definetly a prime number?
			-- `reps' controls how many tests are made (should be 5 - 10)
		require
			i_exists: i /= Void
			reps_positive: reps > 0
		do
			Result := mpz_probab_prime_p (i.item, reps) = 2
		end

	is_probably_prime (i: GMP_INTEGER_REF; reps: INTEGER): BOOLEAN is
			-- Is `i' probably a prime number?
			-- `reps' controls how many tests are made (should be 5 - 10)
		require
			i_exists: i /= Void
		do
			Result := mpz_probab_prime_p (i.item, reps) = 1
		end

	last_root_exact: BOOLEAN
			-- Did last call to `root' return an exact result?

feature -- Implementation

	Prime_test_reps: INTEGER is 10

end -- class GMP_INTEGER_MATH
