indexing
	description: "Math operations on GMP_REAL_REFs"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	GMP_REAL_MATH

inherit
	GMP_REAL_FUNCTIONS
		export
			{NONE} all
		end

	GMP_INTEGER_MATH
		rename
			mul_2_exp as gmp_integer_math_mul_2_exp,
			square_root as gmp_integer_math_square_root
		export
			{NONE} all
		end

feature -- Access

	arc_tangent (d: GMP_REAL_REF): GMP_REAL_REF is
			-- Inverse tangent of `d'.
		require
			d_exists: d /= Void
		local
			x: GMP_REAL_REF
			i, n: INTEGER
		do
			n := d.precision
			x := (d * d) / (One_const + (d * d))
			from
				i := 1
				create Result.make (1.0)
			until
				i > n
			loop
				Result := Result + (get_m2 (i) * (x ^ i))
				i := i + 1
			end
			Result := (Result * d) / (One_const + (d * d))
		end

	cosecant (d: GMP_REAL_REF): GMP_REAL_REF is
		require
			d_exists: d /= Void
			d_not_zero: not d.is_equal (d.zero)
		do
			Result := One_const / sine (d)
		end

	cosine (d: GMP_REAL_REF): GMP_REAL_REF is
		require
			d_exists: d /= Void
		local
			i, j, n: INTEGER
		do
			n := d.precision
			create Result.make (1.0)
			from
				i := 2
				j := -1
			until
				i > n
			loop
				if j = -1 then
					Result := Result - ((d ^ i) / factorial (i).to_gmp_real)
				else
					Result := Result + ((d ^ i) / factorial (i).to_gmp_real)
				end
				i := i + 2
				j := -j
			end
		end

	cotangent (d: GMP_REAL_REF): GMP_REAL_REF is
		require
			d_exists: d /= Void
			d_not_zero: not d.is_equal (d.zero)
		do
			Result := cosine (d) / sine (d)
		end

	div_2_exp (r: GMP_REAL_REF; an_exp: INTEGER): GMP_REAL_REF is
			-- Set `Result' to `r' divided by 2 raised to `an_exp' in the precision of `r'.
		require
			r_exists: r /= Void
			an_exp_not_negative: not (an_exp < 0)
		do
			create Result
			Result.set_precision (r.precision)
			mpf_div_2exp (Result.item, r.item, an_exp) 
		end

	exp (d: GMP_REAL_REF): GMP_REAL_REF is
			-- e ^ `d' in the precision of `d'.
		require
			d_exists: d /= Void
		local
			i, n: INTEGER
		do
			n := d.precision
			create Result.make (1.0)
			from
				i := 1
			until
				i > n
			loop
				Result := Result + ((d ^ i) / factorial (i).to_gmp_real)
				i := i + 1
			end
			Result.set_precision (d.precision)
		end

	hyperbolic_arc_cosine (d: GMP_REAL_REF): GMP_REAL_REF is
			-- Hyperbolic arc sine of `d'.
		require
			d_exists: d /= Void
		do
			Result := log (d + square_root ((d * d) - One_const))
		end

	hyperbolic_arc_sine (d: GMP_REAL_REF): GMP_REAL_REF is
			-- Hyperbolic arc sine of `d'.
		require
			d_exists: d /= Void
		do
			Result := log (d + square_root ((d * d) + One_const))
		end

	hyperbolic_arc_tangent (d: GMP_REAL_REF): GMP_REAL_REF is
			-- Hyperbolic arc sine of `d'.
		require
			d_exists: d /= Void
		do
			Result := One_half_const * log ((One_const + d) / (One_const - d))
		end

	log (d: GMP_REAL_REF): GMP_REAL_REF is
			-- Natural logarithm of `d'.
		require
			d_exists: d /= Void
			d_positive: d > d.zero
		local
			m: GMP_REAL_REF
			i, n: INTEGER
		do
			n := d.precision * 3
			create Result
			m := (d - d.one) / (d + d.one)
			from
				i := 1
			until
				i > n
			loop
				Result := Result + ((One_const / create {GMP_REAL_REF}.make (i)) * (m ^ i))
				i := i + 2
			end
			Result := Result * Two_const
		end

	mul_2_exp (r: GMP_REAL_REF; an_exp: INTEGER): GMP_REAL_REF is
			-- Set `Result' to `r' times 2 raised to `an_exp' in the precision of `r'.
		require
			r_exists: r /= Void
			an_exp_not_negative: not (an_exp < 0)
		do
			create Result
			Result.set_precision (r.precision)
			mpf_mul_2exp (Result.item, r.item, an_exp) 
		end

	secant (d: GMP_REAL_REF): GMP_REAL_REF is
		require
			d_exists: d /= Void
		do
			Result := One_const / cosine (d)
		end

	sine (d: GMP_REAL_REF): GMP_REAL_REF is
		require
			d_exists: d /= Void
		local
			i, j, n: INTEGER
		do
			n := d.precision
			create Result
			Result.copy (d)
			from
				i := 3
				j := -1
			until
				i > n
			loop
				if j = -1 then
					Result := Result - ((d ^ i) / factorial (i).to_gmp_real)
				else
					Result := Result + ((d ^ i) / factorial (i).to_gmp_real)
				end
				i := i + 2
				j := -j
			end
		end

	square_root (r: GMP_REAL_REF): GMP_REAL_REF is
			-- Square rootof `r' in the precision of `r'.
		require
			r_exists: r /= Void
		do
			create Result
			Result.set_precision (r.precision)
			mpf_sqrt (Result.item, r.item)
		end

	tangent (d: GMP_REAL_REF): GMP_REAL_REF is
		require
			d_exists: d /= Void
		do
			Result := sine (d) / cosine (d)
		end

feature {NONE} -- Implementation

	One_const: GMP_REAL_REF is
		once
			create Result.make (1.0)
		end

	One_half_const: GMP_REAL_REF is
		once
			create Result.make (0.5)
		end

	Two_const: GMP_REAL_REF is
		once
			create Result.make (2.0)
		end

	get_m (i: INTEGER): GMP_REAL_REF is
			-- get_m (1) = 2 / 3
			-- get_m (2) = (2 / 3) * (4 / 5)
			-- get_m (3) = (2 / 3) * (4 / 5) * (6 / 7)
			-- get_m (i) = (2 / 3) * (4 / 5) * (6 / 7) * ... * (i * 2, (i * 2) + 1)
		require
--			i_small_enough: i <= 77
		local
			n: INTEGER
			num, den: GMP_REAL_REF
		do
			from
				n := 1
				create num.make (2.0)
				create den.make (3.0)
				create Result.make (1.0)
			until
				n > i  -- if `n' > 77 then `Result' = 0.0
			loop
				Result := (Result * num) / den
				num.copy (num + Two_const)
				den.copy (den + Two_const)
				n := n + 1
			end
		end

	get_m2 (i: INTEGER): GMP_REAL_REF is
			-- get_m (1) = 2 / 3
			-- get_m (2) = (2 / 3) * (4 / 5)
			-- get_m (3) = (2 / 3) * (4 / 5) * (6 / 7)
			-- get_m (i) = (2 / 3) * (4 / 5) * (6 / 7) * ... * (i * 2, (i * 2) + 1)
		local
			n: INTEGER
			den, temp: GMP_INTEGER_REF
		do
			create den.make (1)
			create temp.make (3)
			from
				n := 1
			until
				n > i
			loop
				den.copy (den * temp)
				temp.copy (temp + Two_int)
				n := n + 1
			end
			Result := (create {GMP_RATIONAL_REF}.make ((Two_int ^ i) * factorial (i), den)).to_gmp_real
		end

	Two_int: GMP_INTEGER_REF is
		once
			create Result.make (2)
		end

end -- class GMP_REAL_MATH
