indexing
	description: "Math operations on MPFR_COMPLEX_REFs"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	MPFR_COMPLEX_MATH

inherit
	MPFR_MATH
		rename
			arc_cosecant as arccscnt,
			arc_cosine as arccsn,
			arc_cotangent as arcctngnt,
			arc_hyperbolic_cosecant as archyprblccscnt,
			arc_hyperbolic_cosine as archyprblccsn,
			arc_hyperbolic_cotangent as archyprblcctngnt,
			arc_hyperbolic_secant as arc_hyprblcscnt,
			arc_hyperbolic_tangent as arc_hyprblctagnt,
			arc_secant as arcscnt,
			arc_sine as arcsn,
			arc_tangent as arctngnt,
			cosecant as cscnt,
			cosine as csn,
			cotangent as ctngnt,
			e as gmp_mpfr_math_e,
			euler as gmp_mpfr_math_euler,
			exp as xp,
			gamma as gmm,
			hyperbolic_cosecant as hyprblccscnt,
			hyperbolic_cosine as hyprblc_csn,
			hyperbolic_cotangent as hyprblcctngnt,
			hyperbolic_secant as hyprblcscnt,
			hyperbolic_sine as hyprblc_sn,
			hyperbolic_tangent as hyprblc_tngnt,
			log as lg,
			log_2 as lg2,
			log_10 as lg10,
			log_x as lgx,
			pi as mpfr_math_pi,
			sine as sn,
			secant as scnt,
			square_root as sqrrt,
			tangent as tngnt
		export
			{NONE} arc_tangent_2
		end

feature -- Constants

	e (a_precision: INTEGER): MPFR_COMPLEX_REF is
			-- e with precision `a_precision'
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision)
			                       and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			create Result.make_from_mpfr_real (gmp_mpfr_math_e (a_precision))
		end

	euler, gamma (a_precision: INTEGER): MPFR_COMPLEX_REF is
			-- Euler with precision `a_precision'
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision)
			                       and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			create Result.make_from_mpfr_real (gmp_mpfr_math_euler (a_precision))
		end

	pi (a_precision: INTEGER): MPFR_COMPLEX_REF is
			-- Pi with precision `a_precision'
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision)
			                       and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			create Result.make_from_mpfr_real (mpfr_math_pi (a_precision))
		end

feature -- Exponential

	exp (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Exponential of `z' (e ^ z)
		require
			z_exists: z /= Void
		local
			re: MPFR_REF
		do
			re := xp (z.real)
			create Result.make_from_mpfr_reference (re * csn (z.imaginary), re * sn (z.imaginary))
		end

feature -- Hyperbolic

	hyperbolic_cosecant (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic cosecant
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_one / hyperbolic_sine (i_th, z)
		end

	hyperbolic_cosecant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic cosecant
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_one / hyperbolic_sine_xy (z)
		end

	hyperbolic_cosine (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic cosine
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		local
			re: MPFR_COMPLEX_REF
		do
			re := exp (z.one)
			Result := (power (i_th, re, z) + power (i_th, re, -z))
			Result.real.div_2
			Result.imaginary.div_2
		end

	hyperbolic_cosine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic cosine
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (hyprblc_csn (z.real) * csn (z.imaginary),
			                                        hyprblc_sn (z.real) * sn (z.imaginary))
		end

	hyperbolic_cotangent (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic cotangent
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := hyperbolic_cosine (i_th, z) / hyperbolic_sine (i_th, z)
		end

	hyperbolic_cotangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic cotangent
		require
			z_exists: z /= Void
		local
			x2, y2, div: MPFR_REF
		do
			x2 := z.real + z.real
			y2 := z.imaginary + z.imaginary
			div := hyprblc_csn (x2) - csn (y2)
			create Result.make_from_mpfr_reference (hyprblc_sn (x2) / div, -sn (y2) / div)
		end

	hyperbolic_secant (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic secant
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_one / hyperbolic_cosine (i_th, z)
		end

	hyperbolic_secant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic secant
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_one / hyperbolic_cosine_xy (z)
		end

	hyperbolic_sine (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic sine
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		local
			re: MPFR_COMPLEX_REF
		do
			re := exp (z.one)
			Result := (power (i_th, re, z) - power (i_th, re, -z))
			Result.real.div_2
			Result.imaginary.div_2
		end

	hyperbolic_sine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic sine
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (hyprblc_sn (z.real) * csn (z.imaginary),
			                                        hyprblc_csn (z.real) * sn (z.imaginary))
		end

	hyperbolic_tangent (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic tangent
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := hyperbolic_sine (i_th, z) / hyperbolic_cosine (i_th, z)
		end

	hyperbolic_tangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Hyperbolic tangent
		require
			z_exists: z /= Void
		local
			x2, y2, div: MPFR_REF
		do
			x2 := z.real + z.real
			y2 := z.imaginary + z.imaginary
			div := Mpfr_one / (hyprblc_csn (x2) + csn (y2))
			create Result.make_from_mpfr_reference (hyprblc_sn (x2) * div, sn (y2) * div)
		end

feature -- Inverse hyperbolic

	arc_hyperbolic_cosecant (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic cosecant
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := -log (l_th, square_root (s_th, Mpfr_complex_one + (Mpfr_complex_one / z.square)) + (Mpfr_complex_one / z))
		end

	arc_hyperbolic_cosecant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic cosecant
		require
			z_exists: z /= Void
		do
			Result := -log_xy (square_root_xy (Mpfr_complex_one + (Mpfr_complex_one / z.square)) + (Mpfr_complex_one / z))
		end

	arc_hyperbolic_cosine (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic cosine
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := log (l_th, z + square_root (s_th, z + Mpfr_complex_one) * square_root (s_th, z - Mpfr_complex_one))
		end

	arc_hyperbolic_cosine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic cosine
		require
			z_exists: z /= Void
		do
			Result := log_xy (z + square_root_xy (z + Mpfr_complex_one) * square_root_xy (z - Mpfr_complex_one))
		end

	arc_hyperbolic_cotangent (l_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic cotangent
			-- There are infinitly many
			-- `l_th' selects the logarithm
		require
			z_exists: z /= Void
		do
			Result := (log (l_th, Mpfr_complex_one + (Mpfr_complex_one / z)) -
			           log (l_th, Mpfr_complex_one - (Mpfr_complex_one / z))) * Mpfr_complex_one_half
		end

	arc_hyperbolic_cotangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic cotangent
		require
			z_exists: z /= Void
		do
			Result := (log_xy (Mpfr_complex_one + (Mpfr_complex_one / z)) -
			           log_xy (Mpfr_complex_one - (Mpfr_complex_one / z))) * Mpfr_complex_one_half
		end

	arc_hyperbolic_secant (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic secant
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := arc_hyperbolic_cosine (l_th, s_th, Mpfr_complex_one / z)
		end

	arc_hyperbolic_secant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic secant
		require
			z_exists: z /= Void
		do
			Result := arc_hyperbolic_cosine_xy (Mpfr_complex_one / z)
		end

	arc_hyperbolic_sine (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic sine
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := log (l_th, z + square_root (s_th, z.square + Mpfr_complex_one))
		end

	arc_hyperbolic_sine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic sine
		require
			z_exists: z /= Void
		do
			Result := log_xy (z + square_root_xy (z.square + Mpfr_complex_one))
		end

	arc_hyperbolic_tangent (l_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic tangent
			-- There are infinitly many
			-- `l_th' selects the logarithm
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_one_half * log (l_th, (Mpfr_complex_one + z) / (Mpfr_complex_one - z))
		end

	arc_hyperbolic_tangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc hyperbolic tangent
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_one_half * log_xy ((Mpfr_complex_one + z) / (Mpfr_complex_one - z))
		end

feature -- Inverse trigonometric

	arc_cosecant (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc sine
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := arc_sine (l_th, s_th, z.reciprocal)
		end

	arc_cosecant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc sine
		require
			z_exists: z /= Void
		do
			Result := arc_sine_xy (z.reciprocal)
		end

	arc_cosine (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc cosine
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_i * log (l_th, z + Mpfr_complex_i * square_root (s_th, Mpfr_complex_one - z.square))
		end

	arc_cosine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc cosine
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_i * log_xy (z + Mpfr_complex_i * square_root_xy (Mpfr_complex_one - z.square))
		end

	arc_cotangent (l_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc cotangent
			-- There are infinitly many
			-- `l_th' selects the logarithm
		require
			z_exists: z /= Void
		do
			Result := arc_tangent (l_th, z.reciprocal)
		end

	arc_cotangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc cotangent
		require
			z_exists: z /= Void
		do
			Result := arc_tangent_xy (z.reciprocal)
		end

	arc_secant (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc secant
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := arc_cosine (l_th, s_th, Mpfr_complex_one / z)
		end

	arc_secant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc secant
		require
			z_exists: z /= Void
		do
			Result := arc_cosine_xy (Mpfr_complex_one / z)
		end

	arc_sine (l_th, s_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc sine
			-- There are infinitly many
			-- `l_th' selects the logarithm
			-- `s_th' selects the square_root
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_i * log (l_th, z.multiply_i + square_root (s_th, Mpfr_complex_one - z.square))
		end

	arc_sine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc sine
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_i * log_xy (z.multiply_i + square_root_xy (Mpfr_complex_one - z.square))
		end

	arc_tangent (l_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc sine
			-- There are infinitly many
			-- `l_th' selects the logarithm
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_i_over_2 * log (l_th, (Mpfr_complex_i + z) / (Mpfr_complex_i - z))
		end

	arc_tangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Arc sine
		require
			z_exists: z /= Void
		do
			Result := Mpfr_complex_i_over_2 * log_xy ((Mpfr_complex_i + z) / (Mpfr_complex_i - z))
		end

feature -- Logarithmic

	log (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Natural logarithm of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (lg (z.abs),
			                                        z.arg + (Mpfr_two * create {MPFR_REF}.make (i_th) * mpfr_math_pi (z.precision)))
		end

	log_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Natural logarithm of `z'
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (lg (z.abs), z.arg)
		end

	log_2 (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Base 2 logarithm of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := log (i_th, z) / log (i_th, Mpfr_complex_two)
		end

	log_2_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Base 2 logarithm of `z'
		require
			z_exists: z /= Void
		do
			Result := log_xy (z) / log_xy (Mpfr_complex_two)
		end

	log_10 (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Base 2 logarithm of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := log (i_th, z) / log (i_th, Mpfr_complex_ten)
		end

	log_10_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Base 2 logarithm of `z'
		require
			z_exists: z /= Void
		do
			Result := log_xy (z) / log_xy (Mpfr_complex_ten)
		end

	log_x (i_th: INTEGER; z, x: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Base `x' logarithm of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := log (i_th, z) / log (i_th, x)
		end

	log_x_xy (z, x: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Base `x' logarithm of `z'
		require
			z_exists: z /= Void
		do
			Result := log_xy (z) / log_xy (x)
		end

feature -- Power

	power (i_th: INTEGER; base_z, exp_z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- `base_z' raised to the power `exp_z'
			-- There are infinitly many and `i_th' selects one.
		require
			base_z_exists: base_z /= Void
			exp_z_exists: exp_z /= Void
		do
			Result := exp (log (i_th, base_z) * exp_z)
		end

	power_xy (base_z, exp_z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- `base_z' raised to the power `exp_z'
		require
			base_z_exists: base_z /= Void
			exp_z_exists: exp_z /= Void
		do
			Result := exp (log_xy (base_z) * exp_z)
		end

	power_2_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Raise to the power 3
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (z.real * z.real + z.imaginary * z.imaginary,
			                                        Mpfr_two * z.real * z.imaginary)
		end

	power_3_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Raise to the power 3
		require
			z_exists: z /= Void
		local
			r2, i2: MPFR_REF
		do
			r2 := z.real * z.real
			i2 := z.imaginary * z.imaginary
			create Result.make_from_mpfr_reference (z.real * (r2 - (Mpfr_three * i2)),
			                                        z.imaginary * ((Mpfr_three * r2) - i2))
		end

feature -- Reciporical

	reciporical (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- 1 / z
		require
			z_exists: z /= Void
		local
			den: MPFR_REF
		do
			den := (z.real * z.real) + (z.imaginary * z.imaginary)
			create Result.make_from_mpfr_reference (z.real / den, -z.imaginary / den)
		end

feature -- Root

	root (a_root, i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- There are `a_root' roots of `z' and `i_th' selects one
		require
			a_root_large_enough: a_root >= 2
			i_th_in_range: (1 <= i_th) and (i_th <= a_root)
		local
			theta, n, fr_i_th, fr_r_th: MPFR_REF
		do
			create fr_i_th.make_from_integer (i_th)
			create fr_r_th.make_from_integer (a_root)
			theta := (arctngnt (z.imaginary / z.real) + Mpfr_two * (fr_i_th - Mpfr_one) * mpfr_math_pi (z.precision)) / fr_r_th
			n := hypotenuse (z.real, z.imaginary) ^ (Mpfr_one / fr_r_th)
			create Result.make_from_mpfr_reference (n * csn (theta), n * sn (theta))
		end

	square_root (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Square root of `z'
		require
			z_exists: z /= Void
			valid_i_th: (i_th = 0) or (i_th = 1)
		local
			y: MPFR_REF
		do
			if z.imaginary.is_equal (Mpfr_zero) then
				if z.real < z.real.zero then
					create Result.make_from_mpfr (z.real.zero, sqrrt (z.real.abs))
				else
					create Result.make_from_mpfr (sqrrt (z.real), z.imaginary.zero)
				end
			else
				y := sqrrt ((hypotenuse (z.real, z.imaginary) - z.real) * Mpfr_one_half)
				create Result.make_from_mpfr_reference (z.imaginary / (Mpfr_two * y), y)
			end
			if i_th = 1 then
				Result := -Result
			end
		ensure
			result_exists: Result /= Void
		end

	square_root_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Square root of `z'
		require
			z_exists: z /= Void
		local
			y: MPFR_REF
		do
			if z.imaginary.is_equal (Mpfr_zero) then
				if z.real < z.real.zero then
					create Result.make_from_mpfr (z.real.zero, sqrrt (z.real.abs))
				else
					create Result.make_from_mpfr (sqrrt (z.real), z.imaginary.zero)
				end
			else
				y := sqrrt ((hypotenuse (z.real, z.imaginary) - z.real) * Mpfr_one_half)
				create Result.make_from_mpfr_reference (z.imaginary / (Mpfr_two * y), y)
			end
		ensure
			result_exists: Result /= Void
		end

feature -- Trigonometric

	cosecant (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Cosecant of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
			z_not_zero: not (z.real.is_equal (z.real.zero) and z.imaginary.is_equal (z.imaginary.zero))
		local
			div: MPFR_REF
		do
			Result := sine (i_th, z)
			div := (Result.real * Result.real) + (Result.imaginary * Result.imaginary)
			Result.real.copy (Result.real / div)
			Result.imaginary.copy (-Result.imaginary / div)
		end

	cosecant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Cosecant of `z'
		require
			z_exists: z /= Void
			z_not_zero: not (z.real.is_equal (z.real.zero) and z.imaginary.is_equal (z.imaginary.zero))
		local
			div: MPFR_REF
		do
			Result := sine_xy (z)
			div := (Result.real * Result.real) + (Result.imaginary * Result.imaginary)
			Result.real.copy (Result.real / div)
			Result.imaginary.copy (-Result.imaginary / div)
		end

	cosine (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Coine of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		local
			re, iz: MPFR_COMPLEX_REF
		do
			iz := z.multiply_i
			re := exp (z.one)
			Result := (power (i_th, re, iz) + power (i_th, re, -iz))
			Result.real.div_2
			Result.imaginary.div_2
		end

	cosine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Coine of `z'
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (csn (z.real) * hyprblc_csn (z.imaginary), -sn (z.real) * hyprblc_sn (z.imaginary))
		end

	cotangent (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Coangent of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
			z_not_zero: not (z.real.is_equal (z.real.zero) and z.imaginary.is_equal (z.imaginary.zero))
		do
			Result := cosine (i_th, z) / sine (i_th, z)
		end

	cotangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Tangent of `z'
		require
			z_exists: z /= Void
		local
			x2, y2, div: MPFR_REF
		do
			x2 := z.real + z.real
			y2 := z.imaginary + z.imaginary
			div := hyprblc_csn (y2) - csn (x2)
			create Result.make_from_mpfr_reference (sn (x2) / div, -hyprblc_sn (y2) / div)
		end

	secant (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Secant of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		local
			div: MPFR_REF
		do
			Result := cosine (i_th, z)
			div := (Result.real * Result.real) + (Result.imaginary * Result.imaginary)
			Result.real.copy (Result.real / div)
			Result.imaginary.copy (-Result.imaginary / div)
		end

	secant_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Secant of `z'
		require
			z_exists: z /= Void
		local
			div: MPFR_REF
		do
			Result := cosine_xy (z)
			div := (Result.real * Result.real) + (Result.imaginary * Result.imaginary)
			Result.real.copy (Result.real / div)
			Result.imaginary.copy (-Result.imaginary / div)
		end

	sine (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Sine of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		local
			re, iz: MPFR_COMPLEX_REF
		do
			iz := z.multiply_i
			re := exp (z.one)
			Result := (power (i_th, re, iz) - power (i_th, re, -iz))
			mpfr_swap (Result.real.item, Result.imaginary.item)
			Result.real.div_2
			Result.imaginary.div_2
			Result.imaginary.negate
		end

	sine_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Sine of `z'
		require
			z_exists: z /= Void
		do
			create Result.make_from_mpfr_reference (sn (z.real) * hyprblc_csn (z.imaginary), csn (z.real) * hyprblc_sn (z.imaginary))
		end

	tangent (i_th: INTEGER; z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Tangent of `z'
			-- There are infinitly many and `i_th' selects one.
		require
			z_exists: z /= Void
		do
			Result := sine (i_th, z) / cosine (i_th, z)
		end

	tangent_xy (z: MPFR_COMPLEX_REF): MPFR_COMPLEX_REF is
			-- Tangent of `z'
		require
			z_exists: z /= Void
		local
			x2, y2, div: MPFR_REF
		do
			x2 := z.real + z.real
			y2 := z.imaginary + z.imaginary
			div := csn (x2) + hyprblc_csn (y2)
			create Result.make_from_mpfr_reference (sn (x2) / div, hyprblc_sn (y2) / div)
		end

end -- class MPFR_COMPLEX_MATH
