indexing
	description: "Math operations on MPFR_REFs"
	author: "Christopher John Saunders"
	date: "$Date: 2003/10/17 $"
	revision: "$Revision: 1.0 $"

class
	MPFR_MATH

inherit
	MPFR_FUNCTIONS
		export
			{NONE} all
		end

	MPFR_CONSTANTS

feature -- Constants

	e (a_precision: INTEGER): MPFR_REF is
			-- e with precision `a_precision'
			-- 2.71828182...
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision) and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		local
			a_one: MPFR_REF
		do
			create a_one.make (1.0)
			a_one.set_precision (a_precision)
			Result := exp (a_one)
		end

	euler, gamma (a_precision: INTEGER): MPFR_REF is
			-- Euler with precision `a_precision'
			-- 0.57721566...
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision) and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			create Result
			Result.set_precision (a_precision)
			mpfr_const_euler (Result.item, feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn)
		end

	log_of_2 (a_precision: INTEGER): MPFR_REF is
			-- Natural logarithm of 2 with precision `a_precision'
			-- 0.69314718...
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision) and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			create Result
			Result.set_precision (a_precision)
			mpfr_const_log2 (Result.item, feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn)
		end

	pi (a_precision: INTEGER): MPFR_REF is
			-- Pi with precision `a_precision'
			-- 3.14159265...
		require
			a_precision_in_range: (feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_min <= a_precision) and
			                      (a_precision <= feature {MPFR_PRECISION_CONSTANTS}.Mpfr_prec_max)
		do
			create Result
			Result.set_precision (a_precision)
			mpfr_const_pi (Result.item, feature {MPFR_ROUNDING_MODE_CONSTANTS}.Gmp_rndn)
		end

feature -- Division

	div_2si (op1: MPFR_REF; op2: INTEGER): MPFR_REF is
			-- `op1' / (2 ^ `op2')
		require
			op1_exists: op1 /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op1.precision)
			res := mpfr_div_2si (Result.item, op1.item, op2, op1.rounding_mode)
		end

	div_2ui (op1: MPFR_REF; op2: INTEGER): MPFR_REF is
			-- `op1' / (2 ^ `op2')
		require
			op1_exists: op1 /= Void
			op2_not_negative: not (op2 < 0)
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op1.precision)
			res := mpfr_div_2ui (Result.item, op1.item, op2, op1.rounding_mode)
		end

feature -- Exponential

	exp (op: MPFR_REF): MPFR_REF is
			-- Exponential
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_exp (Result.item, op.item, op.rounding_mode)
		end

feature -- Hyperbolic

	hyperbolic_arc_cosine (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic arc cosine
		require
			op_exists: op /= Void
			not_less_than_one: not (op < Mpfr_one)
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_acosh (Result.item, op.item, op.rounding_mode)
		end

	hyperbolic_arc_sine (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic arc sine
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_asinh (Result.item, op.item, op.rounding_mode)
		end

	hyperbolic_arc_tangent (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic arc tangent
		require
			op_exists: op /= Void
			op_in_domain: op.abs <= Mpfr_one
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_atanh (Result.item, op.item, op.rounding_mode)
		end

	hyperbolic_cosecant (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic cosecant
		require
			op_exists: op /= Void
		do
			Result := Mpfr_one / hyperbolic_sine (op)
		end

	hyperbolic_cosine (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic cosine
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_cosh (Result.item, op.item, op.rounding_mode)
		end

	hyperbolic_cotangent (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic cotangent
		require
			op_exists: op /= Void
		do
			Result := Mpfr_one / hyperbolic_tangent (op)
		end

	hyperbolic_secant (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic secant
		require
			op_exists: op /= Void
		do
			Result := Mpfr_one / hyperbolic_cosine (op)
		end

	hyperbolic_sine (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic cosine
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_sinh (Result.item, op.item, op.rounding_mode)
		end

	hyperbolic_tangent (op: MPFR_REF): MPFR_REF is
			-- Hyperbolic cosine
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_tanh (Result.item, op.item, op.rounding_mode)
		end

feature -- Inverse hyperbolic

	arc_hyperbolic_cosecant (op: MPFR_REF): MPFR_REF is
			-- Arc hyperbolic cosecant
		require
			op_exists: op /= Void
		do
			Result := log (square_root (Mpfr_one + (Mpfr_one / (op * op))) + (Mpfr_one / op))
		end

	arc_hyperbolic_cosine (op: MPFR_REF): MPFR_REF is
			-- Arc hyperbolic cosine
		require
			op_exists: op /= Void
		do
			Result := log (op + square_root ((op * op) - Mpfr_one))
		end

	arc_hyperbolic_cotangent (op: MPFR_REF): MPFR_REF is
			-- Arc hyperbolic cotangent
		require
			op_exists: op /= Void
		do
			Result := Mpfr_one_half * log ((op + Mpfr_one ) / (op - Mpfr_one))
		end

	arc_hyperbolic_secant (op: MPFR_REF): MPFR_REF is
			-- Archyperbolic secant
		require
			op_exists: op /= Void
		do
			Result := log ((Mpfr_one + square_root (Mpfr_one - (op * op))) / op)
		end

	arc_hyperbolic_tangent (op: MPFR_REF): MPFR_REF is
			-- Arc hyperbolic tangent
		require
			op_exists: op /= Void
		do
			Result := Mpfr_one_half * log ((Mpfr_one + op) / (Mpfr_one - op))
		end

feature -- Inverse trigonometric

	arc_cosecant (op: MPFR_REF): MPFR_REF is
			-- Arc cosecant
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_asin (Result.item, (Mpfr_one / op).item, op.rounding_mode)
		end

	arc_cosine (op: MPFR_REF): MPFR_REF is
			-- Arc cosine
		require
			op_exists: op /= Void
			op_in_domain: op.abs <= Mpfr_one
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_acos (Result.item, op.item, op.rounding_mode)
		end

	arc_cotangent (op: MPFR_REF): MPFR_REF is
			-- Arc tangent
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_atan (Result.item, (Mpfr_one / op).item, op.rounding_mode)
		end

	arc_secant (op: MPFR_REF): MPFR_REF is
			-- Arc secant
		require
			op_exists: op /= Void
			op_in_domain: op.abs <= Mpfr_one
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_acos (Result.item, (Mpfr_one / op).item, op.rounding_mode)
		end

	arc_sine (op: MPFR_REF): MPFR_REF is
			-- Arc sine
		require
			op_exists: op /= Void
			op_in_domain: op.abs <= Mpfr_one
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_asin (Result.item, op.item, op.rounding_mode)
		end

	arc_tangent (op: MPFR_REF): MPFR_REF is
			-- Arc tangent
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_atan (Result.item, op.item, op.rounding_mode)
		end

	arc_tangent_2 (y, x: MPFR_REF): MPFR_REF is
			-- Arc tangent of y / x
		require
			x_exists: x /= Void
			y_exists: y /= Void
		do
			Result := arc_tangent (y / x)
		end

feature -- Logarithmic

	log (op: MPFR_REF): MPFR_REF is
			-- Natural logarithm of `op'
		require
			op_exists: op /= Void
			op_in_domain: op >= Mpfr_zero
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_log (Result.item, op.item, op.rounding_mode)
		end

	log_2 (op: MPFR_REF): MPFR_REF is
			-- Base 2 logarithm of `op'
		require
			op_exists: op /= Void
			op_in_domain: op >= Mpfr_zero
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_log2 (Result.item, op.item, op.rounding_mode)
		end

	log_10 (op: MPFR_REF): MPFR_REF is
			-- Base 10 logarithm of `op'
		require
			op_exists: op /= Void
			op_in_domain: op >= Mpfr_zero
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_log10 (Result.item, op.item, op.rounding_mode)
		end

	log_x (op, x: MPFR_REF): MPFR_REF is
			-- Base `x' logarithm of `op'
		require
			op_exists: op /= Void
			op_in_domain: op >= Mpfr_zero
		do
			Result := log (op) / log (x)
		end

feature -- Miscellaneous

	hypotenuse (x, y: MPFR_REF): MPFR_REF is
			-- 
		do
			Result := square_root ((x * x) + (y * y))
		end

feature -- Multiplication

	mul_2si (op1: MPFR_REF; op2: INTEGER): MPFR_REF is
			-- `op1' * (2 ^ `op2')
		require
			op1_exists: op1 /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op1.precision)
			res := mpfr_mul_2si (Result.item, op1.item, op2, op1.rounding_mode)
		end

	mul_2ui (op1: MPFR_REF; op2: INTEGER): MPFR_REF is
			-- `op1' * (2 ^ `op2')
		require
			op1_exists: op1 /= Void
			op2_not_negative: not (op2 < 0)
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op1.precision)
			res := mpfr_mul_2ui (Result.item, op1.item, op2, op1.rounding_mode)
		end

feature -- Root

	square_root (op: MPFR_REF): MPFR_REF is
			-- Square root of `op'
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_sqrt (Result.item, op.item, op.rounding_mode)
		end

feature -- Trigonometric

	cosecant (op: MPFR_REF): MPFR_REF is
			-- Cosecant
		require
			op_exists: op /= Void
		do
			create Result
			Result.set_precision (op.precision)
			Result := Mpfr_one / sine (op)
		end

	cosine (op: MPFR_REF): MPFR_REF is
			-- Arc cosine
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_cos (Result.item, op.item, op.rounding_mode)
		end

	cotangent (op: MPFR_REF): MPFR_REF is
			-- Cotangent
		require
			op_exists: op /= Void
		do
			Result := cosine (op) / sine (op)
		end

	secant (op: MPFR_REF): MPFR_REF is
			-- Secant
		require
			op_exists: op /= Void
		do
			create Result
			Result.set_precision (op.precision)
			Result := Mpfr_one / cosine (op)
		end

	sine (op: MPFR_REF): MPFR_REF is
			-- Sine
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_sin (Result.item, op.item, op.rounding_mode)
		end

	tangent (op: MPFR_REF): MPFR_REF is
			-- Tangent
		require
			op_exists: op /= Void
		local
			res: INTEGER
		do
			create Result
			Result.set_precision (op.precision)
			res := mpfr_tan (Result.item, op.item, op.rounding_mode)
		end

end -- class MPFR_MATH
